import numpy as np
import pandas as pd
import pytest
from scipy.optimize import fsolve
from tdastro.opsim.ztf_opsim import (
    ZTFOpsim,
    calculate_ztf_zero_points,
    create_random_ztf_opsim,
)


def fluxeq(flux, sky=None, gain=None, readnoise=None, nexposure=1, fwhm=None, darkcurrent=None, exptime=None):
    """Define the equation to solve for flux at 5-sigma limit"""

    npix = 2.266 * fwhm**2  # =4*pi*sigma**2 = pi/2/ln2 * FWHM**2
    y = (
        flux**2
        - 25 * flux
        - 25
        * (sky * npix * gain + readnoise**2 * nexposure * npix + darkcurrent * npix * exptime * nexposure)
    )
    return y


def test_calculate_ztf_zero_points():
    """Test the calculation with numerical solution to the maglim function."""

    maglim = 19.0
    sky = 100.0
    fwhm = 1.5
    gain = 1.0
    readnoise = 10.0
    darkcurrent = 0.1
    exptime = 30.0

    root = fsolve(
        lambda x: fluxeq(
            x, sky=sky, gain=gain, readnoise=readnoise, fwhm=fwhm, darkcurrent=darkcurrent, exptime=exptime
        ),
        [100],
    )
    zp_expected = 2.5 * np.log10(root) + maglim

    zp_cal = calculate_ztf_zero_points(
        maglim=maglim,
        sky=sky,
        fwhm=fwhm,
        gain=gain,
        readnoise=readnoise,
        darkcurrent=darkcurrent,
        exptime=exptime,
        nexposure=1,
    )

    assert np.isclose(zp_cal, zp_expected)


def test_ztf_opsim_init():
    """Test initializing ZTFOpsim."""
    opsim_table = create_random_ztf_opsim(100).table
    opsim = ZTFOpsim(table=opsim_table)

    assert opsim.has_columns(["zp_nJy", "obsmjd"])

    # We have all the attributes set at their default values.
    assert opsim.dark_current == 0.0
    assert opsim.gain == 6.2
    assert opsim.pixel_scale == 1.01
    assert opsim.radius == 2.735
    assert opsim.read_noise == 8


def test_create_ztf_opsim_override():
    """Test that we can override the default survey values."""
    opsim_table = create_random_ztf_opsim(100).table

    opsim = ZTFOpsim(
        table=opsim_table,
        dark_current=0.1,
        gain=7.1,
        pixel_scale=0.1,
        radius=1.0,
        read_noise=5.0,
    )

    # We have all the attributes set at their default values.
    assert opsim.dark_current == 0.1
    assert opsim.gain == 7.1
    assert opsim.pixel_scale == 0.1
    assert opsim.radius == 1.0
    assert opsim.read_noise == 5.0


def test_create_ztf_opsim_no_zp():
    """Create an opsim without a zeropoint column."""
    dates = [
        "2020-01-01 12:00:00.000",
        "2020-01-02 12:00:00.000",
        "2020-01-03 12:00:00.000",
        "2020-01-04 12:00:00.000",
        "2020-01-05 12:00:00.000",
    ]
    values = {
        "obsdate": dates,
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
    }

    # We fail if we do not have the other columns needed:
    # "maglim", "sky", "fwhm", "exptime"
    with pytest.raises(ValueError):
        _ = ZTFOpsim(values)

    values["exptime"] = 0.005 * np.ones(5)
    values["maglim"] = 20.0 * np.ones(5)
    values["scibckgnd"] = np.ones(5)
    values["fwhm"] = 2.3 * np.ones(5)
    opsim = ZTFOpsim(values)

    assert opsim.has_columns("zp_nJy")
    assert np.all(opsim["zp_nJy"] >= 0.0)


def test_noise_calculation():
    """Test that the noise calculation is in the right range."""
    mag = np.array([19.0])
    expected_magerr = np.array([0.1])

    flux_nJy = np.power(10.0, -0.4 * (mag - 31.4))
    opsim = ZTFOpsim(
        table=pd.DataFrame(
            {
                "ra": 0.0,
                "dec": 0.0,
                "scibckgnd": 200.0,
                "maglim": 20.0,
                "fwhm": 2.3,
                "exptime": 30.0,
                "obsdate": "2020-01-01 12:00:00.000",
            },
            index=[0],
        )
    )
    fluxerr_nJy = opsim.bandflux_error_point_source(flux_nJy, 0)
    magerr = 1.086 * fluxerr_nJy / flux_nJy

    np.testing.assert_allclose(magerr, expected_magerr, rtol=0.2)



from abc import ABC, abstractmethod

from tdastro.sources.physical_model import PhysicalModel


class PeriodicSource(PhysicalModel, ABC):
    """The base model for periodic sources.

    Parameterized values include:
      * dec - The object's declination in degrees. [from PhysicalModel]
      * distance - The object's luminosity distance in pc. [from PhysicalModel]
      * period - The period of the source, in days.
      * ra - The object's right ascension in degrees. [from PhysicalModel]
      * redshift - The object's redshift. [from PhysicalModel]
      * t0 - The t0 of the zero phase, date. [from PhysicalModel]

    Parameters
    ----------
    period : float
        The period of the source, in days.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(self, period, **kwargs):
        super().__init__(**kwargs)

        # t0 is added in the PhysicalModel constructor.
        self.add_parameter("period", period, **kwargs)

    @abstractmethod
    def _evaluate_phases(self, phases, wavelengths, graph_state, **kwargs):
        """Draw effect-free observations for this object, as a function of phase.

        Parameters
        ----------
        phases : numpy.ndarray
            A length T array of phases, in the range [0, 1].
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths.
        graph_state : GraphState
            An object mapping graph parameters to their values.
        **kwargs : dict, optional
              Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values.
        """
        raise NotImplementedError()

    def compute_flux(self, times, wavelengths, graph_state, **kwargs):
        """Draw effect-free observations for this object.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of rest frame timestamps.
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths.
        graph_state : GraphState
            An object mapping graph parameters to their values.
        **kwargs : dict, optional
           Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values.
        """
        params = self.get_local_params(graph_state)
        period = params["period"]
        phases = (times - params["t0"]) % period / period
        flux_density = self._evaluate_phases(phases, wavelengths, graph_state, **kwargs)

        return flux_density



"""The base models used to specify the TDAstro computation graph.

The computation graph is composed of ParameterizedNodes which compute random variables
(called "model parameters" or "parameters" for short) and encode the dependencies between
them. Model parameters are different from object variables in that they are *not* stored
within the object, but rather programmatically set in an external graph_state dictionary by
sampling the graph. The user does not need to access the internals of graph_state directly,
but rather can treat it as an opaque object to pass around. The dictionary can hold either
individual values (often floats) or arrays of samples.

All model parameters (random variables in the probabilistic graph) must be added using
using the ParameterizedNode.add_parameter() function. This allows the graph to track
which parameters to update and how to set them. Parameter's values can be set from a few
sources:
1) A constant (Example: A given standard deviation for a noise model)
2) A static function or method (which does not have variables that are resampled).
3) The result of evaluating a FunctionNode, which provides a computation using other
   parameters in the graph.
4) The parameters of another ParameterizedNode.

We say that parameter X is dependent on parameter Y if the value of Y is necessary
to compute the value of X. For example if X is set by evaluating a FunctionNode that
uses parameter Y in the computation, X is dependent on Y. The dependencies impose an
ordering of model parameters in the graph. Y must be computed before X.

ParameterNodes provide semantic groupings of individual parameters. For example we may
have a ParameterNode representing the information needed for a Type Ia supernova.
That node's parameters would include the variables needed to evaluate the supernova's
lightcurve. Each of these parameters might depend on parameters in other nodes, such
as those of the host galaxy.

The execution graph is processed by starting at the final node, examining each
model parameter for that node, and recursively proceeding 'up' the graph for any
of its parameters that has a dependency. For example the function.

f(a, b) = x
g(c) = y
h(x, y) = z

would form the graph:

a -\
    x - \
b -/     \
          z
c -- y -- /

where z is the 'bottom' node. Parameters a, b, and c would be at the 'top' of the
graph because they have no dependencies.  Such parameters are set by constants or
static functions.
"""

from functools import partial

import numpy as np

from tdastro.graph_state import GraphState


class ParameterSource:
    """ParameterSource specifies the information about where a ParameterizedNode should
    get the value for a given parameter.

    Attributes
    ----------
    parameter_name : str
        The name of the parameter within the node (short name).
    node_name : str
        The name of the parent node.
    source_type : int
        The type of source as defined by the class variables.
        Default = 0
    value : any
        The information that actually sets the parameter. Either a constant
        or the attribute name of a dependency node.
    dependency : ParameterizedNode or None
        The node on which this parameter is dependent
    allow_gradient : bool
        Allow gradients to be computed at this variable.
        Default = False
    """

    # Class variables for the source enum.
    UNDEFINED = 0
    CONSTANT = 1
    MODEL_PARAMETER = 2
    FUNCTION_NODE = 3
    COMPUTE_OUTPUT = 4

    def __init__(self, parameter_name, source_type=0, node_name=""):
        self.parameter_name = parameter_name
        self.node_name = node_name
        self.source_type = source_type
        self.allow_gradient = False
        self.value = None
        self.dependency = None

    def set_as_constant(self, value, allow_gradient=True):
        """Set the parameter as a constant value.

        Parameters
        ----------
        value : any
            The constant value to use.
        allow_gradient : bool
            Allow a gradient to be computed at this variable.
            Default = True
        """
        if callable(value):
            raise ValueError(f"Using set_as_constant on callable {value}")

        self.source_type = ParameterSource.CONSTANT
        self.allow_gradient = allow_gradient
        self.dependency = None
        self.value = value

    def set_as_parameter(self, dependency, param_name):
        """Set the parameter as a model parameter of another node.  This is
        used for chaining, such as when an object's ra depends on its host's ra.

        Parameters
        ----------
        dependency : ParameterizedNode
            The node in which to access the attribute.
        param_name : str
            The name of the parameter to access.
        """
        self.source_type = ParameterSource.MODEL_PARAMETER
        self.allow_gradient = False
        self.dependency = dependency
        self.value = param_name

    def set_as_function(self, dependency, param_name="function_node_result"):
        """Set the parameter as the result of a FunctionNode (that is not
        the current node).

        Parameters
        ----------
        dependency : ParameterizedNode
            The node in which to access the attribute.
        param_name : str
            The name of where the result is stored in the FunctionNode.
        """
        self.source_type = ParameterSource.FUNCTION_NODE
        self.allow_gradient = False
        self.dependency = dependency
        self.value = param_name

    def set_as_compute_output(self, param_name="function_node_result"):
        """Set the parameter the output of this current node's compute() method.

        This needs to be separate from FUNCTION_NODE type (set_as_function) because
        the sampling function needs to know to call the current node's compute()
        method after the other parameters have been sampled.

        Parameters
        ----------
        dependency : ParameterizedNode
            The node in which to access the attribute.
        param_name : str
            The name of where the result is stored in the FunctionNode.
        """
        self.source_type = ParameterSource.COMPUTE_OUTPUT
        self.allow_gradient = False
        self.value = param_name


class ParameterizedNode:
    """Any model that uses parameters that can be set by constants,
    functions, or other parameterized nodes.

    ParameterizedNodes do not store values directly, but rather provide a recipe
    for how to generate the parameters' values.  The sampled values are read from
    and written to a GraphState object that stores all the parameter values for
    all the nodes.

    Attributes
    ----------
    node_label : str
        An optional human readable identifier (name) for the current node.
    node_string : str
        The full string used to identify a node. This is a combination of the nodes position
        in the graph (if known), node_label (if provided), and class information. This is
        used to access the parameters for this node in the graph_state.
    setters : dict
        A dictionary mapping the parameters' names to information about the setters
        (ParameterSource). The model parameters are stored in the order in which they
        need to be set.
    node_pos : int or None
        A unique ID number for each node in the graph indicating its position.
        Assigned during resampling or set_graph_positions(). This is required to resolve
        naming collisions so we do not overwrite parameters from other nodes.

    Parameters
    ----------
    node_label : str, optional
        An identifier (or name) for the current node.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(self, node_label=None, **kwargs):
        self.setters = {}
        self.node_label = node_label
        self.node_pos = None
        self.node_string = None

        # Give the node a temporary name.
        self._update_node_string()

    def __str__(self):
        """Return the string representation of the node."""
        return self.node_string

    def _update_node_string(self, new_str=None):
        """Update the node's string.

        Parameters
        ----------
        new_str : str, optional
            The new node string. If not provided, the node_string
            is automatically computed from the other node information.
        """
        if self.node_label is not None:
            # If a label is given, just use that. It overrides even the new_str.
            self.node_string = self.node_label
        elif new_str is not None:
            self.node_string = new_str
        else:
            # Otherwise use a combination of the node's class and position.
            pos_string = f"_{self.node_pos}" if self.node_pos is not None else ""
            self.node_string = f"{self.__class__.__name__}{pos_string}"

        # Update the node_name of all node's parameter setters.
        for _, setter_info in self.setters.items():
            setter_info.node_name = self.node_string

    def set_graph_positions(self, seen_nodes=None):
        """Force an update of the graph structure (numbering of each node).

        Parameters
        ----------
        seen_nodes : set, optional
            A set of nodes that have already been processed to prevent infinite loops.
            Caller should not set.
        """
        # Make sure that we do not process the same nodes multiple times.
        if seen_nodes is None:
            seen_nodes = set()
        if self in seen_nodes:
            return
        seen_nodes.add(self)

        # Update the node's position in the graph and its string.
        self.node_pos = len(seen_nodes) - 1
        self._update_node_string()

        # Recursively update any direct dependencies.
        for setter_info in self.setters.values():
            if setter_info.dependency is not None and setter_info.dependency is not self:
                setter_info.dependency.set_graph_positions(seen_nodes)

    def list_params(self):
        """Return a list of this node's parameterized values.

        Returns
        -------
        names : list[str]
            The name of all of the parameterized values for this node.
        """
        return list(self.setters.keys())

    def has_valid_param(self, name):
        """Check whether the node has a given parameterized value and that it is not
        always set to None.

        Parameters
        ----------
        name : str
            The name of the parameter.

        Returns
        -------
        contains : bool
            Whether the node contains a given parameter and it is not always None.
        """
        if name not in self.setters:
            return False

        setter = self.setters[name]
        return not (setter.source_type == ParameterSource.CONSTANT and setter.value is None)

    def get_param(self, graph_state, name, default=None):
        """Get the value of a parameter stored in this node or a default value.

        Note
        ----
        This is an optional helper function that accesses the internals of graph_state
        using the node's information (e.g. its hash value).

        Parameters
        ----------
        graph_state : dict
            The dictionary of graph state information.
        name : str
            The parameter name to query.
        default : any
            The default value to return if the parameter is not in GraphState.

        Returns
        -------
        any
            The parameter value or the default.

        Raises
        ------
        ValueError if graph_state is None.
        """
        if graph_state is None:
            raise ValueError(f"Unable to look up parameter={name}. No graph_state given.")
        if self.node_string in graph_state and name in graph_state[self.node_string]:
            return graph_state[self.node_string][name]
        return default

    def get_local_params(self, graph_state):
        """Get a dictionary of all parameters local to this node.

        Note
        ----
        This is an optional helper function that accesses the internals of graph_state
        using the node's information (e.g. its hash value).

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values.

        Returns
        -------
        result : dict
            A dictionary mapping the parameter name to its value.

        Raises
        ------
        KeyError if no parameters have been set for this node.
        ValueError if graph_state is None.
        """
        if graph_state is None:
            raise ValueError("No graph_state given.")
        return graph_state[self.node_string]

    def set_parameter(self, name, value=None, **kwargs):
        """Set the source of a single *existing* parameter in the ParameterizedNode.

        Parameters within a node are actually a mapping of the parameter name
        to a ParameterSource object that indicates how they are set during sampling.
        Some parameters may be set as a constant value while others may be set by
        evaluating a function that depends on other parameters.

        Notes
        -----
        * Does NOT set an initial value for the model parameter. The user must
          sample the parameters for this to be set.
        * The model parameters are stored in the order in which they are added.

        Parameters
        ----------
        name : str
            The parameter name to add.
        value : any, optional
            The information to use to set the parameter. Can be a constant,
            function, ParameterizedNode, or self.
        **kwargs : dict, optional
           All other keyword arguments, possibly including the parameter setters.

        Raises
        ------
        Raise a KeyError if there is a parameter collision or the parameter
        cannot be found.
        Raise a ValueError if the setter type is not supported.
        """
        # Set the node's position in the graph to None to indicate that the
        # structure might have changed. It needs to be updated with set_graph_positions().
        self.node_pos = None

        # Check for parameter has been added and if so, find the index. All parameters must
        # be added first with "add_parameter()".
        if name not in self.setters:
            raise KeyError(
                f"Tried to set parameter '{name}' that has not been added to node {self.node_string}."
            ) from None

        if value is None and name in kwargs:
            # The value wasn't set, but the name is in kwargs.
            value = kwargs[name]

        if callable(value):
            if "__self__" in value.__dir__() and isinstance(value.__self__, ParameterizedNode):
                # Case 1a: This is a method attached to another ParameterizedNode.
                # Check if this is a getter method, including one that might have been automatically
                # created in add_parameter(). If this is a getter, we access the parameter's sampled
                # values directly to save a function call.
                method_name = value.__name__
                parent = value.__self__
                if method_name in parent.setters:
                    self.setters[name].set_as_parameter(parent, method_name)
                else:
                    raise ValueError(
                        f"Trying to set parameter '{name}' to the {type(parent)}.{method_name}, "
                        f"but unable to find that parameter in class {type(parent)}."
                    )
            else:
                # Case 1b: This is a general function or callable method from another object.
                # We treat it as static (we don't resample the other object) and
                # wrap it in a FunctionNode.
                func_node = FunctionNode(value, **kwargs)
                self.setters[name].set_as_function(func_node)
        elif isinstance(value, FunctionNode):
            # Case 2: We are using the result of a computation of the function node.
            # If the FunctionNode has names outputs that match the variable, use that.
            output_name = name if name in value.outputs else "function_node_result"
            self.setters[name].set_as_function(value, output_name)
        elif isinstance(value, ParameterizedNode):
            # Case 3: We are trying to access a parameter of a ParameterizedNode
            # with the same name.
            if value == self:
                raise ValueError(f"Parameter '{name}' is recursively assigned to self.{name}.")
            if name not in value.setters:
                raise ValueError(f"Parameter '{name}' missing from {str(value)}.")
            self.setters[name].set_as_parameter(value, name)
        else:
            # Case 4: The value is constant (including None).
            self.setters[name].set_as_constant(value)

    def set_allow_gradient(self, name, allow_gradient):
        """Turn on or off the ability to compute a gradient for this variable.

        Parameters
        ----------
        name : str
            The parameter name to modify.
        allow_gradient : bool
            The new setting for allow_gradient.
        """
        self.setters[name].allow_gradient = allow_gradient

    def add_parameter(self, name, value=None, allow_gradient=None, **kwargs):
        """Add a single *new* parameter to the ParameterizedNode.

        Notes
        -----
        * Checks multiple sources in the following order: Manually specified value,
          an entry in kwargs, or None.
        * Does NOT set an initial value for the model parameter. The user must
          sample the parameters for this to be set.
        * The model parameters are stored in the order in which they are added.

        Parameters
        ----------
        name : str
            The parameter name to add.
        value : any, optional
            The information to use to set the parameter. Can be a constant,
            function, ParameterizedNode, or self.
        allow_gradient : bool or None
            Allow gradients to be computed for this variable. If set to None uses the default
            for the setter type (True for constant and False for everything else).
            Default = None
        **kwargs : dict, optional
           All other keyword arguments, possibly including the parameter setters.

        Raises
        ------
        Raise a KeyError if there is a parameter collision or the parameter
        cannot be found.
        """
        # Check for parameter collision and add a place holder value to the 'setters' dictionary.
        if hasattr(self, name) and name not in self.setters:
            raise KeyError(
                f"Parameter name '{name}' conflicts with a predefined model parameter "
                f"or class attribute in {self.node_string}"
            )
        if self.setters.get(name, None) is not None:
            raise KeyError(f"Duplicate parameter set: '{name}' in {self.node_string}")

        # Add an entry for the setter function and fill in the remaining information using
        # set_parameter(). We add an initial (dummy) value here to indicate that this parameter
        # exists and was added via add_parameter().
        self.setters[name] = ParameterSource(
            parameter_name=name,
            source_type=ParameterSource.UNDEFINED,
            node_name=str(self),
        )
        self.set_parameter(name, value, **kwargs)

        # Check if we should override allow_gradient.
        if allow_gradient is not None:
            self.setters[name].allow_gradient = allow_gradient

        # Create a callable getter function with the same name as the parameter.
        # This function allows us to reference the parameter as object.parameter_name
        # for chaining without copying the value. For example, if my_node_1, is a
        # ParameterizedNode with a parameter x, we can do:
        #   my_node_2 = ParameterizedNode(y=my_node_1.x)
        # and my_node_2 will know to use the sampled values of x from my_node_1
        # (as opposed to the setter for x).
        #
        # We override the __self__ and __name__ attributes so it looks like method of
        # this object and the assignment y=my_node_1.x doesn't do a copy of the value.
        def getter(graph_state):
            return graph_state[getter.__self__.node_string][getter.__name__]

        getter.__self__ = self
        getter.__name__ = name
        setattr(self, name, getter)

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Placeholder for a general compute function, which is called at the end
        of the sampling process and can produce derived parameters. This function
        is the main processing step in a FunctionNode.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            Additional function arguments.
        """
        return None

    def _sample_helper(self, graph_state, seen_nodes, rng_info=None):
        """Internal recursive function to sample the model's underlying parameters
        if they are provided by a function or ParameterizedNode. All sampled
        parameters for all nodes are stored in the graph_state dictionary, which is
        modified in-place.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        seen_nodes : dict
            A dictionary mapping nodes strings seen during this sampling run to their object.
            Used to avoid sampling nodes multiple times and to validity check the graph.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.

        Raises
        ------
        Raise a KeyError if the sampling encounters an error with the order of dependencies.
        """
        node_str = str(self)
        if node_str in seen_nodes:
            if seen_nodes[node_str] != self:
                raise ValueError(
                    f"Duplicate node label '{node_str}'. Every node must have a unique label. "
                    "This most often happens when the node_label parameter is set directly."
                )
            return  # Nothing to do
        seen_nodes[node_str] = self

        # Run through each parameter and sample it based on the given recipe.
        # As of Python 3.7 dictionaries are guaranteed to preserve insertion ordering,
        # so this will iterate through model parameters in the order they were inserted.
        any_compute = False
        for name, setter in self.setters.items():
            # Check if we need to sample this parameter's dependency node.
            if setter.dependency is not None and setter.dependency != self:
                setter.dependency._sample_helper(graph_state, seen_nodes, rng_info=rng_info)

            # Set the result from the correct source.
            if setter.source_type == ParameterSource.CONSTANT:
                if graph_state.num_samples == 1:
                    graph_state.set(self.node_string, name, setter.value)
                else:
                    repeated_value = np.array([setter.value] * graph_state.num_samples)
                    graph_state.set(self.node_string, name, repeated_value)
            elif setter.source_type == ParameterSource.MODEL_PARAMETER:
                graph_state.set(
                    self.node_string,
                    name,
                    graph_state[setter.dependency.node_string][setter.value],
                )
            elif setter.source_type == ParameterSource.FUNCTION_NODE:
                graph_state.set(
                    self.node_string,
                    name,
                    graph_state[setter.dependency.node_string][setter.value],
                )
            elif setter.source_type == ParameterSource.COMPUTE_OUTPUT:
                # Computed parameters are set only after all the other (input) parameters.
                any_compute = True
            else:
                raise ValueError(f"Invalid ParameterSource type {setter.source_type}")

        # If this is a function node and the parameters depend on the result of its own computation
        # call the compute function to fill them in.
        if any_compute:
            self.compute(graph_state, rng_info)

    def sample_parameters(self, given_args=None, num_samples=1, rng_info=None):
        """Sample the model's underlying parameters if they are provided by a function
        or ParameterizedNode.

        Parameters
        ----------
        given_args : dict, optional
            A dictionary representing the given arguments for this sample run.
            This can be used as the JAX PyTree for differentiation.
        num_samples : int
            A count of the number of samples to compute.
            Default: 1
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.

        Returns
        -------
        graph_state : GraphState
            A dictionary of dictionaries mapping node->hash, variable_name to either a
            value or array of values. This data structure is modified in place to represent
            the model's state(s).

        Raises
        ------
        Raise a ValueError the sampling encounters a problem with the order of dependencies.
        """
        # If the graph structure has never been set, do that now.
        if self.node_pos is None:
            nodes = set()
            self.set_graph_positions(seen_nodes=nodes)

        # Create space for the results and set all the given_args as fixed parameters.
        results = GraphState(num_samples)
        if given_args is not None:
            results.update(given_args, all_fixed=True)

        # Resample the nodes. All information is stored in the returned results dictionary.
        seen_nodes = {}
        self._sample_helper(results, seen_nodes, rng_info=rng_info)
        return results

    def build_pytree(self, graph_state, partial=None):
        """Build a JAX PyTree representation of the variables in this graph.

        Parameters
        ----------
        graph_state : dict
            A dictionary of dictionaries mapping node->hash, variable_name to value.
            This data structure is modified in place to represent the current state.
        partial : dict
            The partial results so far. This is modified in place by the function.
            A dictionary mapping node name to a dictionary mapping each variable's name
            to its value.
            Default : None

        Returns
        -------
        values : dict
            A dictionary mapping node name to a dictionary mapping each variable's name
            to its value.
        """
        # Check if the node might have incomplete information.
        if self.node_pos is None:
            raise ValueError(
                f"Node {self.node_string} is missing position. You must call "
                "set_graph_positions() before building a pytree."
            )

        # Skip nodes that we have already seen.
        if partial is None:
            partial = {}
        if self.node_string in partial:
            return partial

        # Add new values to the pytree, recursively exploring dependencies.
        partial[self.node_string] = {}
        for name, setter_info in self.setters.items():
            if setter_info.allow_gradient:
                # Anything wth allow_gradient == True goes in the PyTree.
                partial[self.node_string][name] = graph_state[self.node_string][name]
            elif setter_info.dependency is not None:
                # We only recursively check parameters above non-gradient nodes.
                partial = setter_info.dependency.build_pytree(graph_state, partial)
        return partial


class FunctionNode(ParameterizedNode):
    """A class to wrap functions and their argument settings.

    The node can compute the result using a given function (the func
    parameter) or through the compute() method. If no func=None
    then the user must override compute().

    Attributes
    ----------
    func : function or method or partial
        The function to call during an evaluation. If this is None
        you must override the compute() method directly.
    args_names : list
        A list of argument names to pass to the function.
    outputs : list of str
        The output model parameters of this function.

    Parameters
    ----------
    func : function or method
        The function to call during an evaluation.
    node_label : str, optional
        An identifier (or name) for the current node.
    outputs : list of str, optional
        The output model parameters of this function. If None uses
        a single model parameter result.
    fixed_params : dict, optional
        A dictionary mapping a parameter name in the function to its fixed value.
    **kwargs : dict, optional
        Any additional keyword arguments.

    Examples
    --------
    my_func = TDFunc(random.randint, a=1, b=10)
    value1 = my_func()      # Sample from default range
    value2 = my_func(b=20)  # Sample from extended range

    Note
    ----
    All the function's parameters that will be used need to be specified
    in either the default_args dict, object_args list, or as a kwarg in the
    constructor. Arguments cannot be first given during function call.
    For example the following will fail (because b is not defined in the
    constructor):

    my_func = TDFunc(random.randint, a=1)
    value1 = my_func(b=10.0)
    """

    def __init__(self, func, node_label=None, outputs=None, fixed_params=None, **kwargs):
        # We set the function before calling the parent class so we can use
        # the function's name (if needed).
        if fixed_params is not None and len(fixed_params) > 0:
            # Create a partial function with some of the parameters fixed.
            self.func = partial(func, **fixed_params)

            # We need to set the __name__ parameter because it is not preserved by partial.
            self.func.__name__ = func.__name__
        else:
            # Use the function as-is.
            self.func = func

        super().__init__(node_label=node_label, **kwargs)

        # Add all of the parameters from default_args or the kwargs.
        self.arg_names = []
        for key, value in kwargs.items():
            self.arg_names.append(key)
            self.add_parameter(key, value)

        # Add the output arguments.
        if not outputs:
            outputs = ["function_node_result"]
        self.outputs = outputs
        for name in outputs:
            # For output parameters we add a placeholder of None to set up the basic data, such as
            # the getter function and the entry in parameters. Then we change the
            # type to point to own result.
            self.add_parameter(name, None)
            self.setters[name].set_as_compute_output(param_name=name)

    def _non_func(self):
        """This function does nothing. This is used for FunctionNodes where the actual computation
        happens in an overloaded compute() function."""
        pass

    def _update_node_string(self, new_str=None):
        """Update the node's string. A Function node's string includes
        the function name in addition to the class name.
        """
        if new_str is None:
            pos_string = f"_{self.node_pos}" if self.node_pos is not None else ""
            fn_str = f":{self.func.__name__}" if self.func is not None else ""
            new_str = f"{self.__class__.__name__}{fn_str}{pos_string}"
        super()._update_node_string(new_str)

    def _build_inputs(self, graph_state, **kwargs):
        """Build the input arguments for the function.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        args : dict
            A dictionary of input argument to value.
        """
        args = {}
        for key in self.arg_names:
            if key in kwargs:
                args[key] = kwargs[key]
            else:
                args[key] = graph_state[self.node_string][key]
        return args

    def _save_results(self, results, graph_state):
        """Save the results to the graph state.

        Parameters
        ----------
        results : iterable
            The function's results.
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        """
        if len(self.outputs) == 1:
            graph_state.set(self.node_string, self.outputs[0], results)
        else:
            if len(results) != len(self.outputs):
                raise ValueError(
                    f"Incorrect number of results returned by {self.func.__name__}. "
                    f"Expected {len(self.outputs)}, but got {results}."
                )
            for i in range(len(self.outputs)):
                graph_state.set(self.node_string, self.outputs[i], results[i])

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Execute the wrapped function.

        The input arguments are taken from the current graph_state and the outputs
        are written to graph_state.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        results : any
            The result of the computation. This return value is provided so that testing
            functions can easily access the results.

        Raises
        ------
        ValueError is func attribute is None.
        """
        if self.func is None:
            raise ValueError(
                f"The FunctionNode {self.node_string}'s 'func' parameter is None. "
                "You need to either set func or override compute()."
            )

        # Build a dictionary of arguments for the function, call the function, and save
        # the results in the graph state.
        args = self._build_inputs(graph_state, **kwargs)
        results = self.func(**args)
        self._save_results(results, graph_state)
        return results

    def generate(self, given_args=None, num_samples=1, rng_info=None, **kwargs):
        """A helper function that regenerates the parameters for this nodes and the
        ones above it, then returns the the output or this individual node.

        This is used both for testing and for computing JAX gradients.

        Parameters
        ----------
        given_args : dict, optional
            A dictionary representing the given arguments for this sample run.
            This can be used as the JAX PyTree for differentiation.
        num_samples : int
            A count of the number of samples to compute.
            Default: 1
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            Additional function arguments.
        """
        state = self.sample_parameters(given_args, num_samples, rng_info)

        # Get the result(s) of compute from the state object.
        if len(self.outputs) == 1:
            return self.get_param(state, self.outputs[0])

        results = []
        for output_name in self.outputs:
            results.append(self.get_param(state, output_name))
        return results



"""Test magnitude-flux conversion utilities."""

import numpy as np
from tdastro.astro_utils.mag_flux import flux2mag, mag2flux


def test_flux2mag():
    """Test that mag2flux is correct."""
    flux = np.array([3631e9, 1e9, 3631])
    desired_mag = np.array([0, 8.9, 22.5])
    np.testing.assert_allclose(flux2mag(flux), desired_mag, atol=1e-3)


def test_mag2flux():
    """Test that flux2mag is correct."""
    mag = np.array([0, 8.9, 8.9 + 2.5 * 9])
    desired_flux = np.array([3631e9, 1e9, 1])
    np.testing.assert_allclose(mag2flux(mag), desired_flux, rtol=1e-3)


def test_mag2flux2mag():
    """Tesst that mag2flux inverts flux2mag."""
    mag = np.random.uniform(-10, 30, 1024)
    flux = mag2flux(mag)
    mag2 = flux2mag(flux)
    np.testing.assert_allclose(mag, mag2, rtol=1e-10)


def test_flux2mag2flux():
    """Tesst that flux2mag inverts mag2flux."""
    flux = np.random.uniform(1e-3, 1e3, 1024)
    mag = flux2mag(flux)
    flux2 = mag2flux(mag)
    np.testing.assert_allclose(flux, flux2, rtol=1e-10)



import numpy as np
import pytest
from tdastro.effects.basic_effects import ConstantDimming
from tdastro.sources.basic_sources import StaticSource


def test_constant_dimming() -> None:
    """Test that we can sample and create a ConstantDimming."""
    values = np.full((5, 3), 100.0)

    # We can apply the noise.
    effect = ConstantDimming(flux_fraction=0.1)
    values = effect.apply(values, flux_fraction=0.1)
    assert np.all(values == 10.0)

    # We fail if we do not pass in the expected parameters.
    with pytest.raises(ValueError):
        _ = effect.apply(values)


def test_static_source_constant_dimming() -> None:
    """Test that we can sample and create a StaticSource object with constant dimming."""
    model = StaticSource(brightness=10.0, node_label="my_static_source")
    assert len(model.rest_frame_effects) == 0
    assert len(model.obs_frame_effects) == 0

    # We can add the white noise effect.  By default it is a rest frame effect.
    effect = ConstantDimming(flux_fraction=0.1)
    model.add_effect(effect)
    assert len(model.rest_frame_effects) == 1
    assert len(model.obs_frame_effects) == 0

    state = model.sample_parameters()
    times = np.array([1, 2, 3, 4, 5, 10])
    wavelengths = np.array([100.0, 200.0, 300.0])
    values = model.evaluate(times, wavelengths, state)
    assert values.shape == (6, 3)
    assert np.all(values == 1.0)

    # We can add the white noise effect as a observer frame effect instead.
    model2 = StaticSource(brightness=10.0, node_label="my_static_source")
    effect2 = ConstantDimming(flux_fraction=0.5, rest_frame=False)
    model2.add_effect(effect2)
    assert len(model2.rest_frame_effects) == 0
    assert len(model2.obs_frame_effects) == 1

    state2 = model2.sample_parameters()
    values2 = model2.evaluate(times, wavelengths, state2)
    assert values2.shape == (6, 3)
    assert np.all(values2 == 5.0)



from tdastro.sources.physical_model import PhysicalModel


class SNIaHost(PhysicalModel):
    """A SN Ia host galaxy model with a hostmass parameter, more to be added.

    Parameterized values include:
      * dec - The object's declination in degrees. [from PhysicalModel]
      * distance - The object's luminosity distance in pc. [from PhysicalModel]
      * hostmass - The hostmass value.
      * ra - The object's right ascension in degrees. [from PhysicalModel]
      * redshift - The object's redshift. [from PhysicalModel]
      * t0 - The t0 of the zero phase, date. [from PhysicalModel]
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_parameter("hostmass", **kwargs)



from pathlib import Path

from astropy import units as u
from citation_compass import CiteClass

from tdastro.astro_utils.salt2_color_law import SALT2ColorLaw
from tdastro.astro_utils.unit_utils import flam_to_fnu
from tdastro.sources.physical_model import PhysicalModel
from tdastro.utils.bicubic_interp import BicubicInterpolator


class SALT2JaxModel(PhysicalModel, CiteClass):
    """A SALT2 model implemented with JAX for it can use auto-differentiation.

    The model is defined in (Guy J., 2007) as:

    flux(time, wave) = x0 * [M0(time, wave) + x1 * M1(time, wave)] * exp(c * CL(wave))

    where x0, x1, and c are given parameters, M0 is the average spectral sequence,
    M1 is the first compoment to describe variability, and CL is the average color
    correction law.

    We use the formulation in sncosmo where CL is defined such that:

    flux(time, wave) = x0 * [M0(time, wave) + x1 * M1(time, wave)] * 10 ** (-0.4 * c * CL(wave))

    This class is based on the sncosmo implementation at:
    https://github.com/sncosmo/sncosmo/blob/v2.10.1/sncosmo/models.py
    The wrapped sncosmo version in sncosmo_models.py is faster and should be used
    when auto-differentiation is not needed.

    Parameterized values include:
      * c - The SALT2 c parameter.
      * dec - The object's declination in degrees. [from PhysicalModel]
      * distance - The object's luminosity distance in pc. [from PhysicalModel]
      * period - The period of the source, in days.
      * ra - The object's right ascension in degrees. [from PhysicalModel]
      * redshift - The object's redshift. [from PhysicalModel]
      * t0 - The t0 of the zero phase, date. [from PhysicalModel]
      * x0 - The SALT2 x0 parameter.
      * x1 - The SALT2 x0 parameter.

    References
    ----------
    * SALT2: Guy J., 2007 - https://doi.org/10.48550/arXiv.astro-ph/0701828
    * sncosmo - https://zenodo.org/records/14714968

    Attributes
    ----------
    _m0_model : BicubicInterpolator
        The interpolator for the m0 parameter.
    _m1_model : BicubicInterpolator
        The interpolator for the m1 parameter.
    _colorlaw : SALT2ColorLaw
        The data to apply the color law.

    Parameters
    ----------
    x0 : parameter
        The SALT2 x0 parameter.
    x1 : parameter
        The SALT2 x1 parameter.
    c : parameter
        The SALT2 c parameter.
    model_dir : str
        The path for the model file directory.
        Default: ""
    m0_filename : str
        The file name for the m0 model component.
        Default: "salt2_template_0.dat"
    m1_filename : str
        The file name for the m1 model component.
        Default: "salt2_template_1.dat"
    cl_filename : str
        The file name of the color law correction coefficients.
        Default: "salt2_color_correction.dat",
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    # A class variable for the units so we are not computing them each time.
    _FLAM_UNIT = u.erg / u.second / u.cm**2 / u.AA

    def __init__(
        self,
        x0=None,
        x1=None,
        c=None,
        t0=0.0,
        model_dir="",
        m0_filename="salt2_template_0.dat",
        m1_filename="salt2_template_1.dat",
        cl_filename="salt2_color_correction.dat",
        **kwargs,
    ):
        super().__init__(**kwargs)

        # Add the model specific parameters.
        self.add_parameter("x0", x0, **kwargs)
        self.add_parameter("x1", x1, **kwargs)
        self.add_parameter("c", c, **kwargs)

        # Load the data files.
        model_path = Path(model_dir)
        self._m0_model = BicubicInterpolator.from_grid_file(
            model_path / m0_filename,
            scale_factor=1e-12,
        )
        self._m1_model = BicubicInterpolator.from_grid_file(
            model_path / m1_filename,
            scale_factor=1e-12,
        )

        # Use the default color correction values.
        self._colorlaw = SALT2ColorLaw.from_file(model_path / cl_filename)

    def mask_by_time(self, times, graph_state=None):
        """Compute a mask for whether a given time is of interest for a given object.
        For example, a user can use this function to generate a mask to include
        only the observations of interest for a window around the supernova.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        graph_state : GraphState, optional
            An object mapping graph parameters to their values.

        Returns
        -------
        time_mask : numpy.ndarray
            A length T array of Booleans indicating whether the time is of interest.
        """
        if graph_state is None:
            raise ValueError("graph_state needed to compute mask_by_time")

        z = self.get_param(graph_state, "redshift", 0.0)
        if z is None:
            z = 0.0

        t0 = self.get_param(graph_state, "t0", 0.0)
        if t0 is None:
            t0 = 0.0

        # Compute the mask.
        good_times = (times > t0 + -20.0 * (1.0 + z)) & (times < t0 + 50.0 * (1.0 + z))
        return good_times

    def compute_flux(self, phase, wavelengths, graph_state, **kwargs):
        """Draw effect-free observations for this object.

        Parameters
        ----------
        phase : numpy.ndarray
            A length T array of rest frame timestamps.
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.

        **kwargs : dict, optional
           Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        m0_vals = self._m0_model(phase, wavelengths)
        m1_vals = self._m1_model(phase, wavelengths)
        params = self.get_local_params(graph_state)

        flux_density = (
            params["x0"]
            * (m0_vals + params["x1"] * m1_vals)
            * 10.0 ** (-0.4 * self._colorlaw.apply(wavelengths) * params["c"])
        )

        # Convert to the correct units.
        flux_density = flam_to_fnu(
            flux_density,
            wavelengths,
            wave_unit=u.AA,
            flam_unit=self._FLAM_UNIT,
            fnu_unit=u.nJy,
        )
        return flux_density



import numpy as np
from tdastro.math_nodes.single_value_node import SingleVariableNode


def test_single_variable_node():
    """Test that we can create and query a SingleVariableNode."""
    node = SingleVariableNode("A", 10.0)
    assert str(node) == "SingleVariableNode"
    assert node.has_valid_param("A")

    state = node.sample_parameters()
    assert node.get_param(state, "A") == 10


def test_single_variable_node_none():
    """Test that we can create and query a SingleVariableNode with
    a constant value of None."""
    node = SingleVariableNode("A", None)
    assert not node.has_valid_param("A")

    state = node.sample_parameters()
    assert node.get_param(state, "A") is None


def test_single_variable_node_multi_dim():
    """Test that we can create and query a SingleVariableNode with
    a multi-dimensional constant value."""
    # Test with a list.
    node = SingleVariableNode("A", [0, 1, 2])
    state = node.sample_parameters(num_samples=10)
    samples = node.get_param(state, "A")
    assert len(samples) == 10
    for sample in samples:
        assert np.all(sample == [0, 1, 2])

    # Test with a numpy array.
    node = SingleVariableNode("A", np.array([0, 1, 2]))
    state = node.sample_parameters(num_samples=10)
    samples = node.get_param(state, "A")
    assert len(samples) == 10
    for sample in samples:
        assert np.array_equal(sample, np.array([0, 1, 2]))



import numpy as np
import pytest
from tdastro.effects.basic_effects import ConstantDimming
from tdastro.math_nodes.np_random import NumpyRandomFunc
from tdastro.sources.basic_sources import StaticSource, StepSource
from tdastro.sources.multi_source_model import AdditiveMultiSourceModel, RandomMultiSourceModel


def test_additive_multi_source_node() -> None:
    """Test that we can create and evaluate a AdditiveMultiSourceModel."""
    source1 = StaticSource(brightness=10.0, node_label="my_static_source")
    source2 = StepSource(brightness=15.0, t0=1.0, t1=2.0, node_label="my_step_source")
    model = AdditiveMultiSourceModel([source1, source2], node_label="my_multi_source")

    state = model.sample_parameters()
    assert state["my_static_source"]["brightness"] == 10.0
    assert state["my_step_source"]["brightness"] == 15.0

    times = np.array([0.0, 1.5, 3.0])
    wavelengths = np.array([1000.0, 2000.0])

    values = model.evaluate(times, wavelengths, state)
    assert values.shape == (3, 2)
    assert np.allclose(values, [[10.0, 10.0], [25.0, 25.0], [10.0, 10.0]])


def test_additive_multi_source_node_resample() -> None:
    """Test that we can correctly resample a AdditiveMultiSourceModel."""
    ra = NumpyRandomFunc("uniform", low=10.0, high=20.0)
    dec = NumpyRandomFunc("uniform", low=-10.0, high=10.0)

    # Source1 and Source2 share the same ra and dec for each sample, but have
    # different brightness values.
    source1 = StaticSource(
        brightness=NumpyRandomFunc("uniform", low=10.0, high=20.0),
        ra=ra,
        dec=dec,
        node_label="my_static_source",
    )
    source2 = StepSource(
        brightness=NumpyRandomFunc("uniform", low=30.0, high=40.0),
        ra=ra,
        dec=dec,
        t0=1.0,
        t1=2.0,
        node_label="my_step_source",
    )
    model = AdditiveMultiSourceModel([source1, source2], node_label="my_multi_source")

    num_samples = 1000
    rng_info = np.random.default_rng(100)
    state = model.sample_parameters(num_samples=num_samples, rng_info=rng_info)
    assert len(np.unique(state["my_static_source"]["ra"])) > num_samples / 2
    assert len(np.unique(state["my_static_source"]["dec"])) > num_samples / 2
    assert np.allclose(state["my_static_source"]["ra"], state["my_static_source"]["ra"])
    assert np.allclose(state["my_static_source"]["dec"], state["my_static_source"]["dec"])

    brightness_diff = state["my_step_source"]["brightness"] - state["my_static_source"]["brightness"]
    assert np.all(brightness_diff > 5.0)

    # We can evaluate the combined model at all of the times.
    times = np.array([0.0, 1.5, 3.0])
    wavelengths = np.array([1000.0, 2000.0])

    values = model.evaluate(times, wavelengths, state)
    assert values.shape == (1000, 3, 2)

    assert np.allclose(values[:, 0, 0], state["my_static_source"]["brightness"])
    assert np.allclose(values[:, 0, 1], state["my_static_source"]["brightness"])
    assert np.allclose(
        values[:, 1, 0],
        state["my_static_source"]["brightness"] + state["my_step_source"]["brightness"],
    )
    assert np.allclose(
        values[:, 1, 1],
        state["my_static_source"]["brightness"] + state["my_step_source"]["brightness"],
    )
    assert np.allclose(values[:, 2, 0], state["my_static_source"]["brightness"])
    assert np.allclose(values[:, 2, 1], state["my_static_source"]["brightness"])


def test_additive_multi_source_node_redshift() -> None:
    """Test that we handle redshifts separately for each source."""
    source1 = StepSource(brightness=10.0, t0=1.0, t1=3.0, redshift=0.0, node_label="source1")
    source2 = StepSource(brightness=10.0, t0=2.0, t1=4.0, redshift=1.0, node_label="source2")
    model = AdditiveMultiSourceModel([source1, source2], node_label="my_multi_source")

    state = model.sample_parameters()
    times = np.array([0.5, 1.5, 2.5, 3.5, 4.5])
    wavelengths = np.array([1000.0, 2000.0])

    # Compute the expected values:
    # The redshift applied to source1 is a no-op for both time and brightness.
    # The redshift applied to source2 shifts times to [1.25 1.75 2.25 2.75 3.25]
    # and scales up the brightness by a factor of (1 + z) = 2.0.
    contrib1 = np.array([[0.0, 0.0], [10.0, 10.0], [10.0, 10.0], [0.0, 0.0], [0.0, 0.0]])
    contrib2 = np.array([[0.0, 0.0], [0.0, 0.0], [20.0, 20.0], [20.0, 20.0], [20.0, 20.0]])

    values = model.evaluate(times, wavelengths, state)
    assert values.shape == (5, 2)
    assert np.allclose(values, contrib1 + contrib2)


def test_additive_multi_source_node_effects_rest_frame() -> None:
    """Test that we handle rest frame effects separately for each source."""
    source1 = StepSource(brightness=10.0, t0=1.0, t1=3.0, node_label="source1")
    source2 = StepSource(brightness=10.0, t0=2.0, t1=4.0, node_label="source2")
    source1.add_effect(ConstantDimming(flux_fraction=0.5, rest_frame=True))
    source2.add_effect(ConstantDimming(flux_fraction=0.1, rest_frame=True))
    model = AdditiveMultiSourceModel([source1, source2], node_label="my_multi_source")

    # We added the effect to each submodel's rest frame list.
    assert len(source1.rest_frame_effects) == 1
    assert len(source2.rest_frame_effects) == 1
    assert len(source1.obs_frame_effects) == 0
    assert len(source2.obs_frame_effects) == 0
    assert len(model.rest_frame_effects) == 0
    assert len(model.obs_frame_effects) == 0

    state = model.sample_parameters()
    times = np.array([0.5, 1.5, 2.5, 3.5, 4.5])
    wavelengths = np.array([1000.0, 2000.0])
    values = model.evaluate(times, wavelengths, state)
    assert values.shape == (5, 2)

    contrib1 = np.array([[0.0, 0.0], [5.0, 5.0], [5.0, 5.0], [0.0, 0.0], [0.0, 0.0]])
    contrib2 = np.array([[0.0, 0.0], [0.0, 0.0], [1.0, 1.0], [1.0, 1.0], [0.0, 0.0]])
    assert np.allclose(values, contrib1 + contrib2)


def test_additive_multi_source_node_effects_rest_frame_add() -> None:
    """Test that we handle rest frame effects separately for each source."""
    source1 = StepSource(brightness=10.0, t0=1.0, t1=3.0, node_label="source1")
    source2 = StepSource(brightness=10.0, t0=2.0, t1=4.0, node_label="source2")
    model = AdditiveMultiSourceModel([source1, source2], node_label="my_multi_source")
    model.add_effect(ConstantDimming(flux_fraction=0.5, rest_frame=True))

    # We added the effect to each submodel's rest frame list.
    assert len(source1.rest_frame_effects) == 1
    assert len(source2.rest_frame_effects) == 1
    assert len(source1.obs_frame_effects) == 0
    assert len(source2.obs_frame_effects) == 0
    assert len(model.rest_frame_effects) == 0
    assert len(model.obs_frame_effects) == 0

    state = model.sample_parameters()
    times = np.array([0.5, 1.5, 2.5, 3.5, 4.5])
    wavelengths = np.array([1000.0, 2000.0])
    values = model.evaluate(times, wavelengths, state)
    assert values.shape == (5, 2)

    contrib1 = np.array([[0.0, 0.0], [5.0, 5.0], [5.0, 5.0], [0.0, 0.0], [0.0, 0.0]])
    contrib2 = np.array([[0.0, 0.0], [0.0, 0.0], [5.0, 5.0], [5.0, 5.0], [0.0, 0.0]])
    assert np.allclose(values, contrib1 + contrib2)


def test_additive_multi_source_node_effects_obs_frame() -> None:
    """Test that we handle observer frame effects together."""
    source1 = StepSource(brightness=10.0, t0=1.0, t1=3.0, node_label="source1")
    source2 = StepSource(brightness=10.0, t0=2.0, t1=4.0, node_label="source2")
    model = AdditiveMultiSourceModel([source1, source2], node_label="my_multi_source")
    model.add_effect(ConstantDimming(flux_fraction=0.5, rest_frame=False))

    # We added the effect to the joint model's rest frame list.
    assert len(source1.rest_frame_effects) == 0
    assert len(source2.rest_frame_effects) == 0
    assert len(source1.obs_frame_effects) == 0
    assert len(source2.obs_frame_effects) == 0
    assert len(model.rest_frame_effects) == 0
    assert len(model.obs_frame_effects) == 1

    state = model.sample_parameters()
    times = np.array([0.5, 1.5, 2.5, 3.5, 4.5])
    wavelengths = np.array([1000.0, 2000.0])
    values = model.evaluate(times, wavelengths, state)
    assert values.shape == (5, 2)

    contrib1 = np.array([[0.0, 0.0], [5.0, 5.0], [5.0, 5.0], [0.0, 0.0], [0.0, 0.0]])
    contrib2 = np.array([[0.0, 0.0], [0.0, 0.0], [5.0, 5.0], [5.0, 5.0], [0.0, 0.0]])
    assert np.allclose(values, contrib1 + contrib2)


def test_additive_multi_source_node_effects_fail() -> None:
    """Test that we fail if any source includes an observer frame effect."""
    source1 = StepSource(brightness=10.0, t0=1.0, t1=3.0, node_label="source1")
    source2 = StepSource(brightness=10.0, t0=2.0, t1=4.0, node_label="source2")
    source2.add_effect(ConstantDimming(flux_fraction=0.5, rest_frame=False))

    with pytest.raises(ValueError):
        _ = AdditiveMultiSourceModel([source1, source2], node_label="my_multi_source")


def test_random_multi_source_node() -> None:
    """Test that we can create and evaluate a RandomMultiSourceModel."""
    source1 = StaticSource(brightness=10.0, node_label="source1")
    source2 = StaticSource(brightness=15.0, node_label="source2")
    model = RandomMultiSourceModel(
        [source1, source2],
        weights=[0.8, 0.2],
        node_label="my_multi_source",
    )

    state = model.sample_parameters(num_samples=10_000)
    assert np.all(state["source1"]["brightness"] == 10.0)
    assert np.all(state["source2"]["brightness"] == 15.0)

    # We should get approximately 80% of the samples from the first
    # source and 20% from the second source.
    source = np.array(state["my_multi_source"]["selected_source"], dtype=str)
    assert np.all((source == "source1") | (source == "source2"))
    assert np.sum(source == "source1") > 7000
    assert np.sum(source == "source2") > 1000

    # When we evaluate the model, we should get the expected values.
    times = np.array([0.0, 1.5, 3.0])
    wavelengths = np.array([1000.0, 2000.0])
    values = model.evaluate(times, wavelengths, state)

    assert values.shape == (10_000, 3, 2)
    assert np.all((values == 10.0) | (values == 15.0))
    assert np.sum(values == 10.0) > 7000 * 5
    assert np.sum(values == 15.0) > 1000 * 5



import numpy as np
from tdastro.math_nodes.np_random import NumpyRandomFunc
from tdastro.sources.basic_sources import StaticSource
from tdastro.sources.galaxy_models import GaussianGalaxy
from tdastro.sources.multi_source_model import AdditiveMultiSourceModel


def test_gaussian_galaxy() -> None:
    """Test that we can sample and create a StaticSource object."""
    # Create a host galaxy anywhere on the sky.
    host = GaussianGalaxy(
        ra=NumpyRandomFunc("uniform", low=0.0, high=360.0),
        dec=NumpyRandomFunc("uniform", low=-90.0, high=90.0),
        brightness=10.0,
        radius=1.0 / 3600.0,
        node_name="host",
    )

    # We define the position of the source using Gaussian noise from the center of the host galaxy.
    source = StaticSource(
        ra=NumpyRandomFunc("normal", loc=host.ra, scale=host.galaxy_radius_std),
        dec=NumpyRandomFunc("normal", loc=host.dec, scale=host.galaxy_radius_std),
        brightness=100.0,
        node_name="source",
    )

    # Create a combined source / host model.
    combined_source = AdditiveMultiSourceModel(sources=[host, source], ra=source.ra, dec=source.dec)

    # Both RA and dec should be "close" to (but not exactly at) the center of the galaxy.
    state = combined_source.sample_parameters()
    host_ra = host.get_param(state, "ra")
    host_dec = host.get_param(state, "dec")
    source_ra_offset = source.get_param(state, "ra") - host_ra
    assert 0.0 < np.abs(source_ra_offset) < 100.0 / 3600.0

    source_dec_offset = source.get_param(state, "dec") - host_dec
    assert 0.0 < np.abs(source_dec_offset) < 100.0 / 3600.0

    times = np.array([1, 2, 3, 4, 5, 10])
    wavelengths = np.array([100.0, 200.0, 300.0])

    # All the measured fluxes should have some contribution from the background object.
    values = combined_source.evaluate(times, wavelengths)
    assert values.shape == (6, 3)
    assert np.all(values > 100.0)
    assert np.all(values <= 110.0)

    # Check that if we resample the source it will propagate and correctly resample the host.
    # the host's (RA, dec) should change and the source's should still be close.
    state2 = combined_source.sample_parameters()
    assert host_ra != host.get_param(state2, "ra")
    assert host_dec != host.get_param(state2, "dec")

    source_ra_offset2 = host.get_param(state2, "ra") - source.get_param(state2, "ra")
    assert source_ra_offset != source_ra_offset2
    assert 0.0 < np.abs(source_ra_offset2) < 100.0 / 3600.0

    source_dec_offset2 = host.get_param(state2, "dec") - source.get_param(state2, "dec")
    assert source_dec_offset != source_dec_offset2
    assert 0.0 < np.abs(source_ra_offset2) < 100.0 / 3600.0



import astropy.constants as const
import numpy as np
from scipy.integrate import trapezoid
from tdastro.astro_utils.black_body import black_body_luminosity_density_per_solid


def test_bolometric_luminosity():
    """Test that integral of black-body luminosity density is equal to the bolomertric luminosity"""
    speed_of_light = const.c.cgs.value
    bolometric_luminosity = const.L_sun.cgs.value
    radius = const.R_sun.cgs.value
    teff = (bolometric_luminosity / (4 * np.pi * radius**2 * const.sigma_sb.cgs.value)) ** 0.25

    # From 10 to 100,000 Angstroms, in cm
    wavelengths = np.geomspace(100_000, 10, 1001) * 1e-8
    frequencies = speed_of_light / wavelengths

    # Multiply dL_nu / d nu / d Omega by 4pi and integrate over frequency
    luminosity_density = 4 * np.pi * black_body_luminosity_density_per_solid(teff, radius, wavelengths)
    total_luminosity = trapezoid(y=luminosity_density, x=frequencies)

    np.testing.assert_allclose(total_luminosity, bolometric_luminosity, rtol=1e-3)



from pathlib import Path

# Define some global directory paths to use for testing, notebooks, etc.
_TDASTRO_BASE_DIR = Path(__file__).parent.parent.parent
_TDASTRO_BASE_DATA_DIR = _TDASTRO_BASE_DIR / "data"

_TDASTRO_TEST_DIR = _TDASTRO_BASE_DIR / "tests" / "tdastro"
_TDASTRO_TEST_DATA_DIR = _TDASTRO_TEST_DIR / "data"



"""Wrapper classes for calling numpy random number generators."""

from os import urandom

import numpy as np

from tdastro.base_models import FunctionNode


class NumpyRandomFunc(FunctionNode):
    """The base class for numpy random number generators.

    Attributes
    ----------
    func_name : str
        The name of the random function to use.
    _rng : numpy.random._generator.Generator
        This object's random number generator.
    sample_size : tuple
        The shape of the array to generate for each sample. The actual returned value
        will be (num_samples, *size). If an empty tuple will generate a single value per sample.

    Parameters
    ----------
    func_name : str
        The name of the random function to use.
    size : int or tuple, optional
        The shape of the array to generate for each sample. Actual
        returned value will be (num_samples, *size).
        Default: None (single values for each sample)
    seed : int, optional
        The seed to use.

    Notes
    -----
    Since we need to create a new random number generator for this object
    and use that generator's functions, we cannot pass in the function directly.
    Instead we need to pass in the function's name.

    The NumpyRandomFunc node does not support the `choice` function.

    Examples
    --------
    # Create a uniform random number generator between 100.0 and 150.0
    func_node = NumpyRandomFunc("uniform", low=100.0, high=150.0)

    # Create a normal random number generator with mean=5.0 and std=1.0
    func_node = NumpyRandomFunc("normal", loc=5.0, scale=1.0)
    """

    def __init__(self, func_name, size=1, seed=None, **kwargs):
        self.func_name = func_name

        # The node does not support the 'choice' function since it cannot take a list of different
        # lists to use for each sampling run (sample 1 chooses a value from list 1, sample 2 from
        # list 2, etc.).
        if func_name == "choice":
            raise ValueError("The 'choice' function is not supported. Use GivenValueSampler instead.")

        # Convert the given size into a tuple of dimensions or None for a single value per sample.
        if size is None or size == 1:
            self.sample_size = ()
        else:
            # Convert a scalar into a tuple and validate.
            if np.isscalar(size):
                size = (size,)
            if np.any(np.array(size) <= 0):
                raise ValueError("Size of output must be >= 0 in each dimension. Received {size}.")
            self.sample_size = size

        # Get a default random number generator for this object, using the
        # given seed if one is provided.
        if seed is None:
            seed = int.from_bytes(urandom(4), "big")
        self._rng = np.random.default_rng(seed=seed)

        # Check that the function exists in numpy's random number generator library.
        if not hasattr(self._rng, func_name):
            raise ValueError(f"Random function {func_name} does not exist.")
        func = getattr(self._rng, func_name)
        super().__init__(func, **kwargs)

    def set_seed(self, new_seed):
        """Update the random number generator's seed to a given value.

        Parameters
        ----------
        new_seed : int
            The given seed
        """
        self._rng = np.random.default_rng(seed=new_seed)
        self.func = getattr(self._rng, self.func_name)

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Execute the wrapped function.

        The input arguments are taken from the current graph_state and the outputs
        are written to graph_state.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        results : any
            The result of the computation. This return value is provided so that testing
            functions can easily access the results.

        Raises
        ------
        ValueError is func attribute is None.
        """
        args = self._build_inputs(graph_state, **kwargs)

        # If a random number generator is given use that. Otherwise use the default one.
        func = self.func if rng_info is None else getattr(rng_info, self.func_name)

        # Set the size according to the number of samples.
        size_param = (graph_state.num_samples, *self.sample_size)
        if size_param == (1,):
            # If we are generating a single sample use None so it isn't in an array.
            size_param = None

        # Generate the values. Then save and return the results.
        results = func(**args, size=size_param)
        self._save_results(results, graph_state)
        return results



"""The top-level module for survey related data, such as pointing and noise
information. By default the module uses the Rubin OpSim data, but it can be
extended to other survey data as well.
"""

from __future__ import annotations  # "type1 | type2" syntax in Python <3.10

import sqlite3
from pathlib import Path

import numpy as np
import pandas as pd
from scipy.spatial import KDTree

from tdastro import _TDASTRO_BASE_DATA_DIR
from tdastro.astro_utils.mag_flux import mag2flux
from tdastro.astro_utils.noise_model import poisson_bandflux_std
from tdastro.astro_utils.zeropoint import (
    _lsstcam_extinction_coeff,
    _lsstcam_zeropoint_per_sec_zenith,
    flux_electron_zeropoint,
)
from tdastro.consts import GAUSS_EFF_AREA2FWHM_SQ
from tdastro.utils.data_download import download_data_file_if_needed

LSSTCAM_PIXEL_SCALE = 0.2
"""The pixel scale for the LSST camera in arcseconds per pixel."""

_lsstcam_readout_noise = 8.8
"""The standard deviation of the count of readout electrons per pixel for the LSST camera.

The value is from https://smtn-002.lsst.io/v/OPSIM-1171/index.html
"""

_lsstcam_dark_current = 0.2
"""The dark current for the LSST camera in electrons per second per pixel.

The value is from https://smtn-002.lsst.io/v/OPSIM-1171/index.html
"""

_lsstcam_view_radius = 1.75
"""The angular radius of the observation field (in degrees)."""


class OpSim:
    """A wrapper class around the opsim table with cached data for efficiency.

    Parameters
    ----------
    table : dict or pandas.core.frame.DataFrame
        The table with all the OpSim information.
    colmap : dict
        A mapping of short column names to their names in the underlying table.
        Defaults to the Rubin OpSim column names, stored in the class variable
        _opsim_colnames.
    **kwargs : dict
        Additional keyword arguments to pass to the constructor. This includes overrides
        for survey parameters such as:
        - dark_current : The dark current for the camera in electrons per second per pixel.
        - ext_coeff: Mapping of filter names to extinction coefficients.
        - pixel_scale: The pixel scale for the camera in arcseconds per pixel.
        - radius: The angular radius of the observations (in degrees).
        - read_noise: The readout noise for the camera in electrons per pixel.
        - zp_per_sec: Mapping of filter names to zeropoints at zenith.

    Attributes
    ----------
    table : pandas.core.frame.DataFrame
        The table with all the OpSim information.
    colmap : dict
        A mapping of short column names to their names in the underlying table.
    _kd_tree : scipy.spatial.KDTree or None
        A kd_tree of the OpSim pointings for fast spatial queries. We use the scipy
        kd-tree instead of astropy's functions so we can directly control caching.
    pixel_scale : float
        The pixel scale for the camera in arcseconds per pixel.
    read_noise : float
        The readout noise for the camera in electrons per pixel.
    dark_current : float
        The dark current for the camera in electrons per second per pixel.
    ext_coeff : dict or None, optional
        Mapping of filter names to extinction coefficients. Defaults to
        the Rubin OpSim values.
    zp_per_sec : dict or None, optional
        Mapping of filter names to zeropoints at zenith. Defaults to
        the Rubin OpSim values.
    radius : float
        The angular radius of the observations (in degrees).
    """

    _required_names = ["ra", "dec", "time"]

    # Default column names for the Rubin OpSim.
    _default_colnames = {
        "airmass": "airmass",
        "dec": "fieldDec",
        "exptime": "visitExposureTime",
        "filter": "filter",
        "ra": "fieldRA",
        "time": "observationStartMJD",
        "zp": "zp_nJy",  # We add this column to the table
        "seeing": "seeingFwhmEff",
        "skybrightness": "skyBrightness",
        "nexposure": "numExposures",
    }

    # Default survey values.
    _default_survey_values = {
        "dark_current": _lsstcam_dark_current,
        "ext_coeff": _lsstcam_extinction_coeff,
        "pixel_scale": LSSTCAM_PIXEL_SCALE,
        "radius": _lsstcam_view_radius,
        "read_noise": _lsstcam_readout_noise,
        "zp_per_sec": _lsstcam_zeropoint_per_sec_zenith,
    }

    # Class constants for the column names.
    def __init__(
        self,
        table,
        colmap=None,
        **kwargs,
    ):
        if isinstance(table, dict):
            self.table = pd.DataFrame(table)
        else:
            self.table = table

        # Basic validity checking on the column map names.
        self.colmap = self._default_colnames.copy() if colmap is None else colmap
        if "zp" not in self.colmap:
            self.colmap["zp"] = "zp_nJy"
        for name in self._required_names:
            if name not in self.colmap:
                raise KeyError(f"The column name map is missing key={name}")

        # Set the survey-specific parameters using values from either the input or the defaults.
        for key, value in self._default_survey_values.items():
            if key in kwargs:
                setattr(self, key, kwargs[key])
            else:
                setattr(self, key, value)

        # If we are not given zero point data, try to derive it from the other columns.
        if not self.has_columns("zp"):
            self._assign_zero_points()

        # Build the kd-tree.
        self._kd_tree = None
        self._build_kd_tree()

    def __len__(self):
        return len(self.table)

    def __getitem__(self, key):
        """Access the underlying opsim table."""
        # Auto apply the colmap if possible.
        if self.colmap is not None and key in self.colmap:
            return self.table[self.colmap[key]]
        return self.table[key]

    @property
    def columns(self):
        """Get the column names."""
        return self.table.columns

    def get_filters(self):
        """Get the unique filters in the OpSim table."""
        colname = self.colmap.get("filter", "filter")
        if colname not in self.table.columns:
            raise KeyError(f"No filters column found in OpSim table. Expected column: {colname}")
        return np.unique(self.table[colname])

    def has_columns(self, columns):
        """Checks whether OpSim has a column or columns while accounting
        for the colmap.

        Parameters
        ----------
        columns : str or iterable
            The column name or column names to check.

        Returns
        -------
        bool
            True if and only if all the columns are contained in the table.
        """
        if isinstance(columns, str):
            return self.colmap.get(columns, columns) in self.table.columns

        return all(self.colmap.get(col, col) in self.table.columns for col in columns)

    def _build_kd_tree(self):
        """Construct the KD-tree from the opsim table."""
        ra_rad = np.radians(self.table[self.colmap["ra"]].to_numpy())
        dec_rad = np.radians(self.table[self.colmap["dec"]].to_numpy())
        # Convert the pointings to Cartesian coordinates on a unit sphere.
        x = np.cos(dec_rad) * np.cos(ra_rad)
        y = np.cos(dec_rad) * np.sin(ra_rad)
        z = np.sin(dec_rad)
        cart_coords = np.array([x, y, z]).T

        # Construct the kd-tree.
        self._kd_tree = KDTree(cart_coords)

    def _assign_zero_points(self):
        """Assign instrumental zero points in nJy to the OpSim tables."""
        if not self.has_columns(["filter", "airmass", "exptime"]):
            raise ValueError(
                "OpSim does not include the columns needed to derive zero point "
                "information. Required columns: filter, airmass, and exptime."
            )

        zp_values = flux_electron_zeropoint(
            ext_coeff=self.ext_coeff,
            instr_zp_mag=self.zp_per_sec,
            band=self.table[self.colmap.get("filter", "filter")],
            airmass=self.table[self.colmap.get("airmass", "airmass")],
            exptime=self.table[self.colmap.get("exptime", "visitExposureTime")],
        )
        self.add_column(self.colmap.get("zp", "zp"), zp_values, overwrite=True)

    @classmethod
    def from_db(cls, filename, sql_query="SELECT * FROM observations", colmap=None):
        """Create an OpSim object from the data in an opsim db file.

        Parameters
        ----------
        filename : str
            The name of the opsim db file.
        sql_query : str
            The SQL query to use when loading the table.
            Default = "SELECT * FROM observations"
        colmap : dict, optional
            A mapping of short column names to their names in the underlying table.
            If None, uses the default column names for the Rubin OpSim.

        Returns
        -------
        opsim : OpSim
            A table with all of the pointing data.

        Raise
        -----
        FileNotFoundError if the file does not exist.
        ValueError if unable to load the table.
        """
        if colmap is None:
            colmap = cls._default_colnames

        if not Path(filename).is_file():
            raise FileNotFoundError(f"opsim file {filename} not found.")
        con = sqlite3.connect(f"file:{filename}?mode=ro", uri=True)

        # Read the table.
        try:
            opsim = pd.read_sql_query(sql_query, con)
        except Exception:
            raise ValueError("Opsim database read failed.") from None

        # Close the connection.
        con.close()

        return OpSim(opsim, colmap=colmap)

    @classmethod
    def from_url(cls, opsim_url, force_download=False):
        """Construct an OpSim object from a URL to a predefined opsim data file.

        For Rubin OpSim data, you will typically use the latest baseline data set in:
        https://s3df.slac.stanford.edu/data/rubin/sim-data/
        such as:
        https://s3df.slac.stanford.edu/data/rubin/sim-data/sims_featureScheduler_runs3.4/baseline/baseline_v3.4_10yrs.db

        Parameters
        ----------
        opsim_url : str
            The URL to the opsim data file.
        force_download : bool, optional
            If True, the OpSim data will be downloaded even if it already exists locally.
            Default is False.

        Returns
        -------
        opsim : OpSim
            An OpSim object containing the data from the specified URL.
        """
        data_file_name = opsim_url.split("/")[-1]
        data_path = _TDASTRO_BASE_DATA_DIR / "opsim" / data_file_name

        if not download_data_file_if_needed(data_path, opsim_url, force_download=force_download):
            raise RuntimeError(f"Failed to download opsim data from {opsim_url}.")
        return cls.from_db(data_path)

    def add_column(self, colname, values, overwrite=False):
        """Add a column to the current opsim table.

        Parameters
        ----------
        colname : str
            The name of the new column.
        values : int, float, str, list, or numpy.ndarray
            The value(s) to add.
        overwrite : bool
            Overwrite the column is it already exists.
            Default: False
        """
        colname = self.colmap.get(colname, colname)
        if colname in self.table.columns and not overwrite:
            raise KeyError(f"Column {colname} already exists.")

        # If the input is a scalar, turn it into an array of the correct length
        if np.isscalar(values):
            values = np.full((len(self.table)), values)
        self.table[colname] = values

    def write_opsim_table(self, filename, tablename="observations", overwrite=False):
        """Write out an opsim database to a given SQL table.

        Parameters
        ----------
        filename : str
            The name of the opsim db file.
        tablename : str
            The table to which to write.
            Default = "observations"
        overwrite : bool
            Overwrite the existing DB file.
            Default = False

        Raise
        -----
        FileExistsError if the file already exists and overwrite is False.
        """
        if_exists = "replace" if overwrite else "fail"

        con = sqlite3.connect(filename)
        try:
            self.table.to_sql(tablename, con, if_exists=if_exists)
        except Exception:
            raise ValueError("Opsim database write failed.") from None

        con.close()

    def time_bounds(self):
        """Returns the min and max times for all observations in the OpSim.

        Returns
        -------
        t_min, t_max : float, float
            The min and max times for all observations in the OpSim.
        """
        t_min = self.table[self.colmap["time"]].min()
        t_max = self.table[self.colmap["time"]].max()
        return t_min, t_max

    def filter_rows(self, rows):
        """Filter the rows in the OpSim to only include those indices that are provided
        in a list of row indices (integers) or marked True in a mask.

        Parameters
        ----------
        rows : numpy.ndarray
            Either a Boolean array of the same length as the table or list of integer
            row indices to keep.

        Returns
        -------
        new_opsim : OpSim
            A new OpSim object with the reduced rows.
        """
        # Check if we are dealing with a mask of a list of indices.
        rows = np.asarray(rows)
        if rows.dtype == bool:
            if len(rows) != len(self.table):
                raise ValueError(
                    f"Mask length mismatch. Expected {len(self.table)} rows, but found {len(rows)}."
                )
            mask = rows
        else:
            mask = np.full((len(self.table),), False)
            mask[rows] = True

        # Do the actual filtering and generate a new OpSim. This automatically creates
        # the cached data, such as the KD-tree.
        new_table = self.table[mask]
        new_opsim = OpSim(
            new_table,
            colmap=self.colmap,
            ext_coeff=self.ext_coeff,
            zp_per_sec=self.zp_per_sec,
            pixel_scale=self.pixel_scale,
            read_noise=self.read_noise,
            dark_current=self.dark_current,
        )
        return new_opsim

    def is_observed(self, query_ra, query_dec, radius=None):
        """Check if the query point(s) fall within the field of view of any
        pointing in the survey.

        Parameters
        ----------
        query_ra : float or numpy.ndarray
            The query right ascension (in degrees).
        query_dec : float or numpy.ndarray
            The query declination (in degrees).
        radius : float or None, optional
            The angular radius of the observation (in degrees). If None
            uses the default radius for the OpSim.

        Returns
        -------
        seen : bool or list[bool]
            Depending on the input, this is either a single bool to indicate
            whether the query point is observed or a list of bools for an array
            of query points.
        """
        inds = self.range_search(query_ra, query_dec, radius)
        if np.isscalar(query_ra):
            return len(inds) > 0
        return [len(entry) > 0 for entry in inds]

    def range_search(self, query_ra, query_dec, radius=None):
        """Return the indices of the opsim pointings that fall within the field
        of view of the query point(s).

        Parameters
        ----------
        query_ra : float or numpy.ndarray
            The query right ascension (in degrees).
        query_dec : float or numpy.ndarray
            The query declination (in degrees).
        radius : float or None, optional
            The angular radius of the observation (in degrees). If None
            uses the default radius for the OpSim.

        Returns
        -------
        inds : list[int] or list[numpy.ndarray]
            Depending on the input, this is either a list of indices for a single query point
            or a list of arrays (of indices) for an array of query points.
        """
        if query_ra is None or query_dec is None:
            raise ValueError("Query RA and dec must be provided for range search, but got None.")
        if not np.isscalar(query_ra):
            if np.isscalar(query_dec) or len(query_ra) != len(query_dec):
                raise ValueError("Query RA and dec must have the same length.")

            query_dec = np.asarray(query_dec)
            query_ra = np.asarray(query_ra)
            if np.any(query_ra == None) or np.any(query_dec == None):  # noqa: E711
                raise ValueError("Query RA and dec cannot contain None.")
        radius = self.radius if radius is None else radius

        # Transform the query point(s) to 3-d Cartesian coordinate(s).
        ra_rad = np.radians(query_ra)
        dec_rad = np.radians(query_dec)
        x = np.cos(dec_rad) * np.cos(ra_rad)
        y = np.cos(dec_rad) * np.sin(ra_rad)
        z = np.sin(dec_rad)
        cart_query = np.array([x, y, z]).T

        # Adjust the angular radius to a cartesian search radius and perform the search.
        adjusted_radius = 2.0 * np.sin(0.5 * np.radians(radius))
        return self._kd_tree.query_ball_point(cart_query, adjusted_radius)

    def get_observations(self, query_ra, query_dec, radius=None, cols=None):
        """Return the observation information when the query point falls within
        the field of view of a pointing in the survey.

        Parameters
        ----------
        query_ra : float
            The query right ascension (in degrees).
        query_dec : float
            The query declination (in degrees).
        radius : float or None, optional
            The angular radius of the observation (in degrees). If None
            uses the default radius for the OpSim.
        cols : list
            A list of the names of columns to extract. If None returns all the
            columns.

        Returns
        -------
        results : dict
            A dictionary mapping the given column name to a numpy array of values.
        """
        neighbors = self.range_search(query_ra, query_dec, radius)

        results = {}
        if cols is None:
            cols = self.table.columns.to_list()
        for col in cols:
            # Allow the user to specify either the original or mapped column names.
            table_col = self.colmap.get(col, col)
            if table_col not in self.table.columns:
                raise KeyError(f"Unrecognized column name {table_col}")
            results[col] = self.table[table_col][neighbors].to_numpy()
        return results

    def bandflux_error_point_source(self, bandflux, index):
        """Compute observational bandflux error for a point source

        Parameters
        ----------
        bandflux : array_like of float
            Band bandflux of the point source in nJy.
        index : array_like of int
            The index of the observation in the OpSim table.

        Returns
        -------
        flux_err : array_like of float
            Simulated bandflux noise in nJy.
        """
        observations = self.table.iloc[index]

        # By the effective FWHM definition, see
        # https://smtn-002.lsst.io/v/OPSIM-1171/index.html
        # We need it in pixel^2
        footprint = (
            GAUSS_EFF_AREA2FWHM_SQ
            * (observations[self.colmap.get("seeing", "seeingFwhmEff")] / self.pixel_scale) ** 2
        )

        zp = observations[self.colmap.get("zp", "zp_nJy")]

        # Table value is in mag/arcsec^2
        sky_njy_angular = mag2flux(observations[self.colmap.get("skybrightness", "skyBrightness")])
        # We need electrons per pixel^2
        sky = sky_njy_angular * self.pixel_scale**2 / zp

        return poisson_bandflux_std(
            bandflux,
            total_exposure_time=observations[self.colmap.get("exptime", "visitExposureTime")],
            exposure_count=observations[self.colmap.get("nexposure", "numExposures")],
            footprint=footprint,
            sky=sky,
            zp=zp,
            readout_noise=self.read_noise,
            dark_current=self.dark_current,
        )


def create_random_opsim(num_obs, seed=None):
    """Create a random OpSim pointings drawn uniformly from (RA, dec).

    Parameters
    ----------
    num_obs : int
        The size of the OpSim to generate.
    seed : int
        The seed to used for random number generation. If None then
        uses a default random number generator.
        Default: None

    Returns
    -------
    opsim_data : OpSim
        The OpSim data structure.
    seed : int, optional
        The seed for the random number generator.
    """
    if num_obs <= 0:
        raise ValueError("Number of observations must be greater than zero.")

    rng = np.random.default_rng() if seed is None else np.random.default_rng(seed=seed)

    # Generate the (RA, dec) pairs uniformly on the surface of a sphere.
    ra = np.degrees(rng.uniform(0.0, 2.0 * np.pi, size=num_obs))
    dec = np.degrees(np.arccos(2.0 * rng.uniform(0.0, 1.0, size=num_obs) - 1.0) - (np.pi / 2.0))

    # Generate the information needed to compute zeropoint.
    airmass = rng.uniform(1.3, 1.7, size=num_obs)
    filter = rng.choice(["u", "g", "r", "i", "z", "y"], size=num_obs)

    input_data = {
        "observationStartMJD": 0.05 * np.arange(num_obs),
        "fieldRA": ra,
        "fieldDec": dec,
        "airmass": airmass,
        "filter": filter,
        "visitExposureTime": 29.0 * np.ones(num_obs),
    }

    opsim = OpSim(input_data)
    return opsim


def opsim_add_random_data(opsim_data, colname, min_val=0.0, max_val=1.0):
    """Add a column composed of random uniform data. Used for testing.

    Parameters
    ----------
    opsim_data : OpSim
        The OpSim data structure to modify.
    colname : str
        The name of the new column to add.
    min_val : float
        The minimum value of the uniform range.
        Default: 0.0
    max_val : float
        The maximum value of the uniform range.
        Default: 1.0
    """
    values = np.random.uniform(low=min_val, high=max_val, size=len(opsim_data))
    opsim_data.add_column(colname, values)


def oversample_opsim(
    opsim: OpSim,
    *,
    pointing: tuple[float, float] = (200, -50),
    search_radius: float = 1.75,
    delta_t: float = 0.01,
    time_range: tuple[float | None, float | None] = (None, None),
    bands: list[str] | None = None,
    strategy: str = "darkest_sky",
):
    """Single-pointing oversampled OpSim table.

    It includes observations for a single pointing only,
    but with very high time resolution. The observations
    would alternate between the bands.

    Parameters
    ----------
    opsim : OpSim
        The OpSim table to oversample.
    pointing : tuple of RA and Dec in degrees
        The pointing to use for the oversampled table.
    search_radius : float, optional
        The search radius for the oversampled table in degrees.
        The default is the half of the LSST's field of view.
    delta_t : float, optional
        The time between observations in days.
    time_range : tuple or floats or Nones, optional
        The start and end times of the observations in MJD.
        None means to use the minimum (maximum) time in
        all the observations found for the given pointing.
        Time is being samples as np.arange(*time_range, delta_t).
    bands : list of str or None, optional
        The list of bands to include in the oversampled table.
        The default is to include all bands found for the given pointing.
    strategy : str, optional
        The strategy to select prototype observations.
        - "darkest_sky" selects the observations with the minimal sky brightness
          (maximum "skyBrightness" value) in each band. This is the default.
        - "random" selects the observations randomly. Fixed seed is used.

    """
    ra, dec = pointing
    observations = opsim.table.iloc[opsim.range_search(ra, dec, search_radius)]
    if len(observations) == 0:
        raise ValueError("No observations found for the given pointing.")

    time_min, time_max = time_range
    if time_min is None:
        time_min = np.min(observations["observationStartMJD"])
    if time_max is None:
        time_max = np.max(observations["observationStartMJD"])
    if time_min >= time_max:
        raise ValueError(f"Invalid time_range: start > end: {time_min} > {time_max}")

    uniq_bands = np.unique(observations["filter"])
    if bands is None:
        bands = uniq_bands
    elif not set(bands).issubset(uniq_bands):
        raise ValueError(f"Invalid bands: {bands}")

    new_times = np.arange(time_min, time_max, delta_t)
    n = len(new_times)
    if n < len(bands):
        raise ValueError("Not enough time points to cover all bands.")

    new_table = pd.DataFrame(
        {
            # Just in case, to not have confusion with the original table
            "observationId": opsim.table["observationId"].max() + 1 + np.arange(n),
            "observationStartMJD": new_times,
            "fieldRA": ra,
            "fieldDec": dec,
            "filter": np.tile(bands, n // len(bands)),
        }
    )
    other_columns = [column for column in observations.columns if column not in new_table.columns]

    if strategy == "darkest_sky":
        for band in bands:
            # MAXimum magnitude is MINimum brightness (darkest sky)
            idxmax = observations["skyBrightness"][observations["filter"] == band].idxmax()
            idx = new_table.index[new_table["filter"] == band]
            darkest_sky_obs = pd.DataFrame.from_records([observations.loc[idxmax]] * idx.size, index=idx)
            new_table.loc[idx, other_columns] = darkest_sky_obs[other_columns]
    elif strategy == "random":
        rng = np.random.default_rng(0)
        for band in bands:
            single_band_obs = observations[observations["filter"] == band]
            idx = new_table.index[new_table["filter"] == band]
            random_obs = single_band_obs.sample(idx.size, replace=True, random_state=rng).set_index(idx)
            new_table.loc[idx, other_columns] = random_obs[other_columns]
    else:
        raise ValueError(f"Invalid strategy: {strategy}")

    return OpSim(
        new_table,
        colmap=opsim.colmap,
        pixel_scale=opsim.pixel_scale,
        read_noise=opsim.read_noise,
        dark_current=opsim.dark_current,
    )



import logging
from pathlib import Path

import numpy as np
import sncosmo
from astropy import units as u
from scipy.interpolate import interp1d
from tdastro.astro_utils.noise_model import apply_noise
from tdastro.astro_utils.passbands import PassbandGroup
from tdastro.astro_utils.snia_utils import (
    DistModFromRedshift,
    HostmassX1Func,
    X0FromDistMod,
    num_snia_per_redshift_bin,
)
from tdastro.astro_utils.unit_utils import flam_to_fnu, fnu_to_flam
from tdastro.math_nodes.np_random import NumpyRandomFunc
from tdastro.math_nodes.scipy_random import SamplePDF
from tdastro.sources.sncomso_models import SncosmoWrapperModel
from tdastro.sources.snia_host import SNIaHost

logger = logging.getLogger(__name__)


def construct_snia_source(oversampled_observations, zpdf):
    """Create a SNIA source/host pair with characteristics from an OpSim.

    Parameters
    ----------
    oversampled_observations : OpSim
        The opsim data to use.
    zpdf : interp1d
        The PDF for the redshift.

    Returns
    -------
    source : PhysicalModel
        The PhysicalModel to sample.
    """
    logger.info("Creating the source model.")

    # Get the range in which t0 can occur.
    t_min = oversampled_observations["observationStartMJD"].min()
    t_max = oversampled_observations["observationStartMJD"].max()

    # TODO: Extract the ra and dec center from the opsim in case that changes.
    # Currently we are relying on the fact the test opsim has all pointings
    # at (0.0, 0.0).

    # Create a host galaxy.
    host = SNIaHost(
        ra=NumpyRandomFunc("uniform", low=-0.5, high=0.5),  # all pointings RA = 0.0
        dec=NumpyRandomFunc("uniform", low=-0.5, high=0.5),  # all pointings Dec = 0.0
        hostmass=NumpyRandomFunc("uniform", low=7, high=12),
        redshift=SamplePDF(zpdf),
        node_label="host",
    )

    distmod_func = DistModFromRedshift(host.redshift, H0=73.0, Omega_m=0.3)
    x1_func = HostmassX1Func(host.hostmass)
    c_func = NumpyRandomFunc("normal", loc=0, scale=0.02)
    m_abs_func = NumpyRandomFunc("normal", loc=-19.3, scale=0.1)
    x0_func = X0FromDistMod(
        distmod=distmod_func,
        x1=x1_func,
        c=c_func,
        alpha=0.14,
        beta=3.1,
        m_abs=m_abs_func,
        node_label="x0_func",
    )

    sncosmo_modelname = "salt3"
    source = SncosmoWrapperModel(
        sncosmo_modelname,
        t0=NumpyRandomFunc("uniform", low=t_min, high=t_max),
        x0=x0_func,
        x1=x1_func,
        c=c_func,
        ra=NumpyRandomFunc("normal", loc=host.ra, scale=0.01),
        dec=NumpyRandomFunc("normal", loc=host.dec, scale=0.01),
        redshift=host.redshift,
        node_label="source",
    )
    return source


def load_and_register_passband(passbands_dir, to_use):
    """Load the passband from files and register with sncosmo.

    Parameters
    ----------
    passbands_dir : str
        The directory containing the passband files to use.
    to_use : list
        A list of the passbands to use.
        Example: ["g", "r"]

    Returns
    -------
    passbands : PassbandGroup
        The loaded and processed PassbandGroup.
    """
    passbands_dir = Path(passbands_dir)
    passband_list = []
    for band in to_use:
        file_path = passbands_dir / "LSST" / f"{band}.dat"
        passband_list.append({"filter_name": band, "table_path": file_path})
        logger.info(f"Loading band {band} from {file_path}")

    if len(passband_list) == 0:
        raise ValueError("No passbands being loaded.")

    # Do the actual loading and processing.
    passbands = PassbandGroup(
        passband_list,
        survey="LSST",
        units="nm",
        trim_quantile=0.001,
        delta_wave=1,
    )

    # Register sncosmo bandpasses
    for f, passband in passbands.passbands.items():
        sncosmo_bandpass = sncosmo.Bandpass(*passband.processed_transmission_table.T, name=f"tdastro_{f}")
        sncosmo.register(sncosmo_bandpass, force=True)

    return passbands


def draw_single_random_sn(
    source,
    opsim,
    passbands,
    state=None,
    rng_info=None,
):
    """Process a single random SN realization.

    Parameters
    ----------
    source : PhysicalModel
        The PhysicalModel to use for the flux computation.
    opsim : OpSim
        The OpSim for the simulations
    passbands : PassbandGroup
        The passbands to use in generating the observations.
    state : GraphState
        The sample values to use. If None resamples the state.
    rng_info : numpy.random._generator.Generator, optional
        A given numpy random number generator to use for this computation. If not
        provided, the function uses the default random number generator.

    Returns
    -------
    res : dict
        A dictionary of useful information about the run.
    """
    if state is None:
        state = source.sample_parameters(rng_info=rng_info)

    # Extract some important parameters that we need to use.
    ra = state["source"]["ra"]
    dec = state["source"]["dec"]
    t0 = state["source"]["t0"]
    z = state["source"]["redshift"]

    # Compute the rest wavelength information and save it.
    wave_obs = passbands.waves
    wavelengths_rest = wave_obs / (1.0 + z)
    res = {"wavelengths_rest": wavelengths_rest}

    # Find the times at which this source is seen.
    obs_index = np.array(opsim.range_search(ra, dec, radius=1.75))

    # Update obs_index to only include observations within SN lifespan
    phase_obs = opsim["time"][obs_index] - t0
    obs_index = obs_index[(phase_obs > -20 * (1.0 + z)) & (phase_obs < 50 * (1.0 + z))]

    # Extract the timing and filter information for those observations, changing the
    # match band names in passbands object.
    times = opsim["time"][obs_index].to_numpy()
    if len(times) == 0:
        logger.warning(f"No overlap time in opsim for (ra,dec)=({ra:.2f},{dec:.2f})")
    res["times"] = times

    filters = opsim["filter"][obs_index].to_numpy(str)
    filters = np.char.add("LSST_", filters)
    res["filters"] = filters

    # Compute the fluxes over all wavelengths.
    flux_nJy = source.evaluate(times, wave_obs, graph_state=state, rng_info=rng_info)
    res["flux_nJy"] = flux_nJy
    res["flux_flam"] = fnu_to_flam(
        flux_nJy,
        wave_obs,
        wave_unit=u.AA,
        flam_unit=u.erg / u.second / u.cm**2 / u.AA,
        fnu_unit=u.nJy,
    )
    res["flux_fnu"] = flux_nJy

    # Compute the band_flixes over just the given filters.
    bandfluxes_perfect = source.get_band_fluxes(passbands, times, filters, state)
    res["bandfluxes_perfect"] = bandfluxes_perfect

    bandfluxes_error = opsim.bandflux_error_point_source(bandfluxes_perfect, obs_index)
    res["bandfluxes_error"] = bandfluxes_error
    res["bandfluxes"] = apply_noise(bandfluxes_perfect, bandfluxes_error, rng=None)

    res["state"] = state

    return res


def run_snia_end2end(
    oversampled_observations,
    passbands_dir,
    solid_angle=0.0001,
    nsample=1,
    check_sncosmo=False,
    rng_info=None,
):
    """Test that we can sample and create SN Ia simulation using the salt3 model.

    Parameters
    ----------
    oversampled_observations : OpSim
        The opsim data to use.
    passbands_dir : str
        The name of the directory holding the passband information.
    solid_angle : float
        Solid angle for calculating number of SN.
    nsample : int
        The number of samples to test.
        Default:  1
    check_sncosmo : bool
        Run the simulation a second time directly with sncosmo and compare the answers.
        This should only be turned on for testing.
        Default: False
    rng_info : numpy.random._generator.Generator, optional
        A given numpy random number generator to use for this computation. If not
        provided, the function uses the default random number generator.

    Returns
    -------
    res_list : dict
        A dictionary of lists of sampling and result information.
    passbands : PassbandGroup
        The passbands used.
    """
    if rng_info is None:
        rng_info = np.random.default_rng()

    # Compute the distribution from which to sample the redshift.
    zmin = 0.1
    zmax = 0.4
    H0 = 70.0
    Omega_m = 0.3
    nsn, z = num_snia_per_redshift_bin(zmin, zmax, 100, H0=H0, Omega_m=Omega_m)
    zpdf = interp1d(z, nsn, bounds_error=False, fill_value=0)

    # Calculate nsample using SN Ia rate model
    if nsample is None and solid_angle is not None:
        nsn, _ = num_snia_per_redshift_bin(
            zmin, zmax, znbins=1, solid_angle=solid_angle, H0=H0, Omega_m=Omega_m
        )
        nsample = int(nsn[0])  # Since znbins=1 there is only one bin
        print(f"Drawing {nsample} samples from redshift {zmin} to {zmax}.")

    source = construct_snia_source(oversampled_observations, zpdf)
    passbands = load_and_register_passband(passbands_dir, to_use=["g", "r"])

    logger.info(f"Sampling {nsample} states.")
    sample_states = source.sample_parameters(num_samples=nsample)

    res_list = []
    for i in range(0, nsample):
        current_state = sample_states.extract_single_sample(i)
        res = draw_single_random_sn(
            source,
            opsim=oversampled_observations,
            passbands=passbands,
            state=current_state,
            rng_info=rng_info,
        )

        # Copy out important parameter values.
        p = {}
        for parname in ["t0", "x0", "x1", "c", "redshift", "ra", "dec"]:
            p[parname] = float(current_state["source"][parname])
        p["hostmass"] = current_state["host.hostmass"]
        p["distmod"] = current_state["x0_func.distmod"]
        res["parameter_values"] = p

        if check_sncosmo:
            saltpars = {"x0": p["x0"], "x1": p["x1"], "c": p["c"], "z": p["redshift"], "t0": p["t0"]}
            model = sncosmo.Model("salt3")
            model.update(saltpars)
            wave = passbands.waves
            time = res["times"]
            filters = res["filters"]

            flux_sncosmo = model.flux(time, wave)
            fnu_sncosmo = flam_to_fnu(
                flux_sncosmo,
                wave,
                wave_unit=u.AA,
                flam_unit=u.erg / u.second / u.cm**2 / u.AA,
                fnu_unit=u.nJy,
            )
            np.testing.assert_allclose(res["flux_nJy"], fnu_sncosmo, atol=1e-8, rtol=1e-6)
            np.testing.assert_allclose(res["flux_flam"], flux_sncosmo, atol=1e-30, rtol=1e-5)

            # Skip test for negative fluxes
            if np.all(flux_sncosmo > 0):
                sncosmo_band_names = np.char.add("tdastro_", filters)
                bandflux_sncosmo = model.bandflux(sncosmo_band_names, time, zpsys="ab", zp=8.9 + 2.5 * 9)
                np.testing.assert_allclose(res["bandfluxes_perfect"], bandflux_sncosmo, rtol=1e-1)

        res_list.append(res)

    return res_list, passbands



import numpy as np
import pytest
from tdastro.utils.wave_extrapolate import (
    ConstantExtrapolation,
    ExponentialDecay,
    LastValueExtrapolation,
    LinearDecay,
    WaveExtrapolationModel,
)


def test_wave_extrapolation_model():
    """Test the base class for the extrapolation methods."""
    # Create an instance of the base class
    extrapolator = WaveExtrapolationModel()

    # Test that the __call__ method returns a zero matrix
    last_wave = 1000.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_waves = np.array([1000.0, 1025.0, 1050.0, 1200.0])
    result = extrapolator(last_wave, last_flux, query_waves)
    expected_result = np.zeros((3, 4))
    np.testing.assert_allclose(result, expected_result)


def test_constant_extrapolation():
    """Test that the constant extrapolation function works."""
    # Create an instance of the ConstantExtrapolation class
    extrapolator = ConstantExtrapolation(value=100.0)

    # Test extrapolation past the last valid point.
    last_wave = 1000.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_waves = np.array([1000.0, 1025.0, 1050.0, 1200.0])
    expected_flux = np.full((3, 4), 100.0)
    result = extrapolator(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)

    # Test that we fail if the value is not positive
    with pytest.raises(ValueError):
        _ = ConstantExtrapolation(value=-1)


def test_last_value_extrapolation():
    """Test that the last value extrapolation function works."""
    # Create an instance of the LastValueExtrapolation class
    extrapolator = LastValueExtrapolation()

    # Test extrapolation past the last valid point.
    last_wave = 1000.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_waves = np.array([1000.0, 1025.0, 1050.0, 1200.0])
    expected_flux = np.array(
        [
            [100.0, 100.0, 100.0, 100.0],
            [200.0, 200.0, 200.0, 200.0],
            [300.0, 300.0, 300.0, 300.0],
        ]
    )
    result = extrapolator(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)


def test_linear_decay_extrapolate():
    """Test that the linear decay function works."""
    decay_width = 100.0
    extrapolator = LinearDecay(decay_width=decay_width)

    # Test extrapolation past the last valid point.
    last_wave = 1000.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_waves = np.array([1000.0, 1025.0, 1050.0, 1075.0, 1100.0, 1150.0])
    expected_flux = np.array(
        [
            [100.0, 75.0, 50.0, 25.0, 0.0, 0.0],
            [200.0, 150.0, 100.0, 50.0, 0.0, 0.0],
            [300.0, 225.0, 150.0, 75.0, 0.0, 0.0],
        ]
    )
    result = extrapolator(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)

    # Test extrapolation before the first valid point.
    query_waves = np.array([1000.0, 975.0, 950.0, 925.0, 900.0, 850.0])
    result = extrapolator(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)

    # Test that we fail if the decay width is not positive
    with pytest.raises(ValueError):
        _ = LinearDecay(decay_width=-1.0)


def test_exponential_decay_extrapolate():
    """Test that the exponential decay function works."""
    extrapolator = ExponentialDecay(rate=0.1)

    # Test extrapolation past the last valid point.
    last_wave = 1000.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_waves = np.array([1000.0, 1025.0, 1050.0, 1075.0, 1100.0, 1150.0])

    t0_flux = 100.0 * np.exp([-0.0, -2.5, -5.0, -7.5, -10.0, -15.0])
    expected_flux = np.vstack((t0_flux, t0_flux * 2, t0_flux * 3))
    result = extrapolator(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)

    # Test extrapolation before the first valid point.
    query_waves = np.array([1000.0, 975.0, 950.0, 925.0, 900.0, 850.0])
    result = extrapolator(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)

    # Test that we fail if the decay rate is not positive
    with pytest.raises(ValueError):
        _ = ExponentialDecay(rate=-0.1)



"""Test the utility functions for computing AGN parameters. These are mostly change
detection tests to make sure the results match the code from SNANA:
https://github.com/RickKessler/SNANA/blob/master/src/gensed_AGN.py
"""

import numpy as np
import pytest
from tdastro.consts import M_SUN_G
from tdastro.math_nodes.np_random import NumpyRandomFunc
from tdastro.sources.agn import AGN, sample_damped_random_walk


def test_agn_compute_critical_accretion_rate():
    """Test that we can compute the accretion rate from mass of the black hole."""
    # This is a change detection test to make sure the results match previous code.

    # No BH = no accretion.
    assert AGN.compute_critical_accretion_rate(0.0) == 0.0

    # For a 1000 solar mass black hole, the accretion rate should be 1.4e21 g/s.
    assert AGN.compute_critical_accretion_rate(1000.0 * M_SUN_G) == pytest.approx(1.4e21)

    # Test that the operation is vectorized.
    num_suns = np.array([0.5, 1.0, 10.0, 50.0])
    masses = M_SUN_G * num_suns
    expected = 1.4e18 * num_suns
    assert np.allclose(AGN.compute_critical_accretion_rate(masses), expected)


def test_agn_black_hole_accretion_rate():
    """Test that we can compute the black hole's accretion rate."""
    # This is a change detection test to make sure the results match previous code.

    # No accretion => no blackhole accretion.
    assert AGN.compute_blackhole_accretion_rate(0.0, 0.5) == 0.0
    assert AGN.compute_blackhole_accretion_rate(0.0, 1.0) == 0.0
    assert AGN.compute_blackhole_accretion_rate(0.0, 2.0) == 0.0

    # For a total accretion rate of 1.4e21 g/s and different ratios,
    # compute the black hole accretion rate.
    assert AGN.compute_blackhole_accretion_rate(1.4e21, 1.0) == pytest.approx(1.4e21)
    assert AGN.compute_blackhole_accretion_rate(1.4e21, 0.5) == pytest.approx(7.0e20)
    assert AGN.compute_blackhole_accretion_rate(1.4e21, 2.0) == pytest.approx(2.8e21)

    # Test that the operation is vectorized.
    rates = np.array([0.5, 1.0, 10.0, 50.0])
    ratios = np.array([1.0, 1.0, 2.0, 2.0])
    assert np.allclose(AGN.compute_blackhole_accretion_rate(rates, ratios), rates * ratios)


def test_agn_compute_bolometric_luminosity():
    """Test that we can compute the bolometric luminosity from mass of the black hole."""
    # This is a change detection test to make sure the results match previous code.

    # No BH = no luminosity.
    assert AGN.compute_bolometric_luminosity(1.0, 0.0) == 0.0

    # For a 100 solar mass black hole and 1.0 ratio, the bolometric luminosity should be 1.26e40
    assert AGN.compute_bolometric_luminosity(1.0, 100.0 * M_SUN_G) == pytest.approx(1.26e40)

    # For a 100 solar mass black hole and 0.5 ratio, the bolometric luminosity should be 6.3e39
    assert AGN.compute_bolometric_luminosity(0.5, 100.0 * M_SUN_G) == pytest.approx(6.3e39)

    # Test that the operation is vectorized.
    num_suns = np.array([0.5, 1.0, 10.0, 50.0])
    ratios = np.array([1.0, 1.0, 2.0, 2.0])
    masses = M_SUN_G * num_suns

    expected = 1.26e38 * num_suns * ratios
    assert np.allclose(AGN.compute_bolometric_luminosity(ratios, masses), expected)


def test_agn_compute_mag_i():
    """Test that we can compute the i band magnitude from the bolometric luminosity."""
    # This is a change detection test to make sure the results match previous code.

    assert AGN.compute_mag_i(1.0) == pytest.approx(90.0)

    # Test that the operation is vectorized.
    l_bol = np.array([0.5, 1.0, 10.0, 50.0])
    expected = np.array([90.75257499, 90.0, 87.5, 85.75257499])
    assert np.allclose(AGN.compute_mag_i(l_bol), expected)


def test_agn_compute_r_0():
    """Test that we can compute an AGN's r_0 from its r_in."""
    # This is a change detection test to make sure the results match previous code.

    assert AGN.compute_r_0(1.0) == pytest.approx(1.361111111111111)

    # Test that the operation is vectorized.
    r_in = np.array([0.5, 1.0, 10.0, 50.0])
    expected = 1.361111111111111 * r_in
    assert np.allclose(AGN.compute_r_0(r_in), expected)


def test_agn_compute_temp_at_r_0():
    """Test that we can compute the effective temperature at r0."""
    # This is a change detection test to make sure the results match previous code.

    assert AGN.compute_temp_at_r_0(1000.0, 100.0, 5.0) == pytest.approx(0.28248581706023856)


def test_compute_structure_function_at_inf():
    """Test that we can compute the structure function at infinity."""
    # This is a change detection test to make sure the results match previous code.
    sf_inf = AGN.compute_structure_function_at_inf(np.array([1000.0, 2000.0, 3000.0, 4000.0]))
    expected = np.array([8.83866626e-05, 6.34152002e-05, 5.22210566e-05, 4.54988060e-05])
    assert np.allclose(sf_inf, expected)


def test_compute_tau_v_drw():
    """Test that we can compute the DRW tau."""
    # This is a change detection test to make sure the results match previous code.
    tau_v = AGN.compute_tau_v_drw(np.array([1000.0, 2000.0, 3000.0, 4000.0]))
    expected = np.array([4546.21322992, 5114.75576753, 5479.74582924, 5754.39937337])
    assert np.allclose(tau_v, expected)


def test_sample_damped_random_walk():
    """Test that we can sample from a damped random walk."""
    rng = np.random.default_rng(100)

    tau_v = np.array([1.0, 1.0, 10.0, 10.0, 100.0])
    sf_inf = np.array([0.1, 0.2, 0.1, 0.2, 0.1])
    times = np.arange(0.0, 20.0, 1.0)
    delta_m = sample_damped_random_walk(times, tau_v, sf_inf, 0.0, rng=rng)
    assert delta_m.shape == (20, 5)

    # The mean of delta_m for each wavelength should be around sf_inf.
    assert np.allclose(np.mean(delta_m, axis=0), sf_inf, atol=0.1)

    # We fail with non-monotonically increasing times.
    with pytest.raises(ValueError):
        _ = sample_damped_random_walk(np.array([0.0, 1.0, 0.0]), tau_v, sf_inf, 0.0)

    # We fail with mismatched tau_v and sf_inf.
    with pytest.raises(ValueError):
        _ = sample_damped_random_walk(times, tau_v[:3], sf_inf, 0.0)


def test_create_agn():
    """Test that we can create an AGN object and the derived parameters are correct."""
    # Select the black hole mass uniformly between 1e9 and 2e9 solar masses.
    bh_mass_sampler = NumpyRandomFunc("uniform", low=1e9 * M_SUN_G, high=2e9 * M_SUN_G, seed=100)
    agn_node = AGN(
        t0=0.0,
        blackhole_mass=bh_mass_sampler,
        edd_ratio=0.9,
        node_label="AGN",
    )
    state = agn_node.sample_parameters(num_samples=10_000)

    # Check that the parameters are within the expected ranges.
    bh_masses = state["AGN"]["blackhole_mass"]
    assert np.all(bh_masses >= 1e9 * M_SUN_G)
    assert np.all(bh_masses <= 2e9 * M_SUN_G)
    assert np.unique(bh_masses).size > 1_000

    assert np.allclose(state["AGN"]["critical_accretion_rate"], 1.4e18 * bh_masses / M_SUN_G)
    assert np.allclose(
        state["AGN"]["blackhole_accretion_rate"], 0.9 * state["AGN"]["critical_accretion_rate"]
    )
    assert np.allclose(
        state["AGN"]["bolometric_luminosity"], AGN.compute_bolometric_luminosity(0.9, bh_masses)
    )
    assert np.allclose(state["AGN"]["mag_i"], AGN.compute_mag_i(state["AGN"]["bolometric_luminosity"]))

    # Check that we can sample from the AGN.
    single_state = agn_node.sample_parameters(num_samples=1)
    times = np.array([10.0, 20.0, 30.0, 40.0, 50.0])
    wavelengths = np.array([1000.0, 2000.0, 3000.0, 4000.0])

    fluxes = agn_node.compute_flux(times, wavelengths, single_state)
    assert fluxes.shape == (5, 4)
    assert np.all(fluxes >= 0.0)



import os

import jax.numpy as jnp
import numpy as np
import pytest
from tdastro.utils.bicubic_interp import (
    BicubicAxis,
    BicubicInterpolator,
    _kernel_value,
    expand_to_cross_products,
)


def test_expand_to_cross_products():
    """Test that we can create cross product arrays."""
    x_vals = np.arange(5, 10)
    y_vals = np.arange(-3, 3)
    num_x = len(x_vals)
    num_y = len(y_vals)

    # Manually fill in the expected results.
    expected_x = np.zeros(num_x * num_y)
    expected_y = np.zeros(num_x * num_y)
    count = 0
    for row in range(num_x):
        for col in range(num_y):
            expected_x[count] = x_vals[row]
            expected_y[count] = y_vals[col]
            count += 1

    all_x, all_y = expand_to_cross_products(x_vals, y_vals)
    assert np.array_equal(all_x, expected_x)
    assert np.array_equal(all_y, expected_y)


def test_bicubic_range():
    """Test that we can create and query a BicubicAxis object."""
    xvals = BicubicAxis([0.1 * i for i in range(100)])
    assert xvals.min_val == 0.0
    assert xvals.max_val == 9.9
    assert xvals.num_vals == 100
    assert xvals.step_size == pytest.approx(0.1)

    # Test that we can run out_of_bounds
    values = [-1.0, 2.0, 11.0, 5.0, -0.1, 3.5, 4.0]
    expected = [True, False, True, False, True, False, False]
    is_out = xvals.out_of_bounds(values)
    assert np.array_equal(is_out, expected)

    # Test that we can run find_indices
    ix = xvals.find_indices(values)
    assert np.array_equal(ix, [0, 20, 98, 50, 0, 35, 40])

    # Run a larger find_indices search.
    n_x = len(xvals)
    x_q = jnp.asarray([0.04 * i - 0.1 for i in range(400)])
    ix = xvals.find_indices(x_q)
    assert jnp.all(ix >= 0)
    assert jnp.all(ix <= n_x - 2)
    assert jnp.all((ix == n_x - 2) | (x_q < xvals.values[ix + 1]))
    assert jnp.all((ix == 0) | (xvals.values[ix] <= x_q))

    # Test that we fail for invalid BicubicAxis range objects.
    with pytest.raises(ValueError):
        _ = BicubicAxis([0.0, 0.1])  # Too few points
    with pytest.raises(ValueError):
        _ = BicubicAxis([0.0, 0.2, 0.1])  # Not sorted


def test_bicubic_range_non_regular():
    """Test that we can create and query a BicubicAxis object
    without regular steps in the values."""
    xvals = BicubicAxis([0.1, 0.2, 0.4, 0.5, 0.6, 0.65, 0.7])
    assert xvals.min_val == 0.1
    assert xvals.max_val == 0.7
    assert xvals.num_vals == 7
    assert xvals.step_size is None

    # Test that we can run out_of_bounds
    values = [0.05, 0.15, 0.25, 0.8, 0.5]
    expected = [True, False, False, True, False]
    is_out = xvals.out_of_bounds(values)
    assert np.array_equal(is_out, expected)

    # Test that we can run find_indices
    ix = xvals.find_indices(values)
    assert np.array_equal(ix, [0, 0, 1, 5, 3])


def test_bicubic_interpolator_simple():
    """Test that we can create and query a bicubic interpolator."""
    x_vals = np.arange(1.0, 5.0, 1.0)
    y_vals = np.arange(0.0, 3.0, 0.5)
    z_vals = np.full((len(x_vals), len(y_vals)), 1.0)
    interp_obj = BicubicInterpolator(x_vals, y_vals, z_vals)

    x_queries = np.arange(0.5, 5.5, 0.5)
    y_queries = np.arange(-0.5, 3.5, 0.5)
    results = interp_obj(x_queries, y_queries)

    # Everything within the grid should be 1.0 and everything
    # outside the grid should be 0.0.
    for i, x_q in enumerate(x_queries):
        for j, y_q in enumerate(y_queries):
            if (x_q < 1.0) or (x_q > 4.0) or (y_q < 0.0) or (y_q > 2.5):
                assert np.abs(results[i, j]) < 1e-8
            else:
                assert np.abs(results[i, j] - 1.0) < 1e-8


def test_bicubic_interpolator():
    """Test that we can create and query a bicubic interpolator."""
    x_vals = np.arange(1.0, 5.0, 1.0)
    num_x = len(x_vals)

    y_vals = np.arange(-1.0, 2.0, 0.5)
    num_y = len(y_vals)

    # Make a wavey plane of values.
    z_vals = np.empty((num_x, num_y))
    for i in range(num_x):
        for j in range(num_y):
            if j % 2 == 0:
                z_vals[i, j] = i + 0.2 * j
            else:
                z_vals[i, j] = i - 0.1 * j

    # Pick some query points.
    x_q = jnp.asarray([-0.5, 1.0, 1.5, 2.0, 2.1, 3.4, 4.0, 5.5])
    y_q = jnp.asarray([-1.5, 0.0, 0.5, 1.1, 2.0, 3.5])

    # Expected values (as computed by the sncosmo version)
    expected = np.array(
        [
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.4, -0.3, 0.54, 0.0, 0.0],
            [0.0, 0.9, 0.2, 1.04, 0.0, 0.0],
            [0.0, 1.4, 0.7, 1.54, 0.0, 0.0],
            [0.0, 1.5, 0.8, 1.64, 0.0, 0.0],
            [0.0, 2.8, 2.1, 2.94, 0.0, 0.0],
            [0.0, 3.4, 2.7, 3.54, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        ]
    )

    # Do the JAX interpolation.
    interp_obj = BicubicInterpolator(x_vals, y_vals, z_vals)
    res_jax = interp_obj(x_q, y_q)

    for i in range(len(x_q)):
        for j in range(len(y_q)):
            if jnp.abs(res_jax[i, j] - expected[i, j]) > 1e-5:
                print(f"TEST {i}, {j} ({x_q[i]}, {y_q[j]}) = {res_jax[i, j]} vs {expected[i, j]}")

    assert jnp.allclose(expected, res_jax, rtol=1e-3, atol=1e-4)


def test_bicubic_interpolator2():
    """Test that we can create and query a bicubic interpolatorusing other values."""
    x_vals = [0.25 * i for i in range(100)]
    y_vals = [j / 8.0 for j in range(320)]
    z_vals = jnp.reshape(jnp.arange(len(x_vals) * len(y_vals)), (len(x_vals), len(y_vals)))

    x_q = jnp.asarray([-0.5, 0.0, 1.0, 22.2, 49.0, 50.5])
    y_q = jnp.asarray([-1.0, 5.0, 13.4, 21.5, 41.0])

    # Expected values (as computed by the sncosmo version)
    expected = np.array(
        [
            [0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 40.0, 107.19999695, 172.0, 0.0],
            [0.0, 1320.0, 1387.19999695, 1452.0, 0.0],
            [0.0, 28456.00097656, 28523.20097351, 28588.00097656, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0],
        ]
    )

    # Do the JAX interpolation.
    interp_obj = BicubicInterpolator(x_vals, y_vals, z_vals)
    res_jax = interp_obj(x_q, y_q)

    assert jnp.allclose(expected, res_jax, rtol=1e-5, atol=1e-5)


def test_kernel_values():
    """Test that we can evaluate the kernel with the JAX function
    and get the same results as with the original function."""
    input = jnp.asarray([-1.5, -1.1, 0.0, 0.5, 1.0, 1.5, 2.0, 2.5])
    expected = jnp.asarray([-0.0625, -0.0405, 1.0, 0.5625, 0.0, -0.0625, 0.0, 0.0])
    all_results = _kernel_value(input)
    assert jnp.allclose(expected, all_results)


def test_bad_bicubic_interpolator():
    """Test that we catch failure cases for creating a bicubic interpolator."""
    good_x = [1.0, 1.1, 1.2]
    good_y = [1.0, 1.1, 1.2]
    good_z = [[1.0, 1.1, 1.2], [1.0, 1.1, 1.2], [1.0, 1.1, 1.2]]
    _ = BicubicInterpolator(good_x, good_y, good_z)

    # No too few x or y
    with pytest.raises(ValueError):
        _ = BicubicInterpolator([1.0], good_y, good_z)
    with pytest.raises(ValueError):
        _ = BicubicInterpolator(good_x, [1.0], good_z)

    # The x and y values must be in sorted order.
    z = np.full((3, 4), 1.0)
    _ = BicubicInterpolator([1.0, 2.0, 3.0], [0.0, 0.5, 1.0, 1.5], z)
    with pytest.raises(ValueError):
        _ = BicubicInterpolator([2.0, 1.0, 3.0], [0.0, 0.5, 1.0, 1.5], z)
    with pytest.raises(ValueError):
        _ = BicubicInterpolator([1.0, 2.0, 3.0], [0.0, 2.5, 1.0, 1.5], z)

    # Sizes must match z.
    with pytest.raises(ValueError):
        _ = BicubicInterpolator([1.0, 2.0, 3.0, 4.0], [0.0, 2.5, 1.0, 1.5], z)
    with pytest.raises(ValueError):
        _ = BicubicInterpolator([1.0, 2.0, 3.0], [0.0, 2.5, 1.0], z)


def test_load_bicubic_interp_from_file(test_data_dir):
    """Test loading a bicubic interpolator object from a file."""
    filename = os.path.join(test_data_dir, "truncated-salt2-h17/salt2_template_0.dat")
    test_obj = BicubicInterpolator.from_grid_file(filename)

    assert len(test_obj.x_vals) == 26
    assert len(test_obj.y_vals) == 401
    assert test_obj.z_vals.shape == (26, 401)



import numpy as np
import numpy.typing as npt


def poisson_bandflux_std(
    bandflux: npt.ArrayLike,
    *,
    total_exposure_time: npt.ArrayLike,
    exposure_count: npt.ArrayLike,
    footprint: npt.ArrayLike,
    sky: npt.ArrayLike,
    zp: npt.ArrayLike,
    readout_noise: npt.ArrayLike,
    dark_current: npt.ArrayLike,
) -> npt.ArrayLike:
    """Simulate photon noise for bandflux measurements.

    Parameters
    ----------
    bandflux : array_like of float
        Source bandflux in energy units, e.g. nJy.
    total_exposure_time : array_like of float
        Total exposure time of all observation, in time units
        (e.g. seconds).
    exposure_count : array_like of int
        Number of exposures in the observation.
    sky : array_like of float
        Sky background per unit angular area,
        in the units of electrons / pixel^2.
    footprint : array_like of float
        Point spread function effective area, in pixel^2.
    zp : array_like of float
        Zero point bandflux for the observation, i.e. bandflux
        giving a single electron during the total exposure time.
        Units are the same as the input bandflux over electron,
        e.g. nJy / electron.
    readout_noise : array_like of float
        Standard deviation of the readout electrons per pixel per exposure.
    dark_current : array_like of float
        Mean dark current electrons per pixel per unit time.

    Returns
    -------
    array_like
        Simulated bandflux noise, in the same units as the input bandflux.

    Notes
    -----

    1. We do not specify units for the input parameters, but they
    should be consistent with each other.

    2. Here we assume that the sky and source photon noises follow
    Poisson statistics in the limit of large number of photons,
    e.g. they are both considered to be normal distributed with
    variance equal to the number of photons. Readout noise is
    assumed to be Poisson distributed with variance (squared mean)
    equal to the square of the given value. Dark current is assumed
    to be Poisson distributed with variance (squared mean) equal
    to the product of the given value and the exposure time.
    The output is Poisson standard deviation of the sum of all
    these noises converted to the flux units.
    """
    # Get variances, in electrons^2
    source_variance = bandflux / zp
    sky_variance = sky * footprint
    readout_variance = readout_noise**2 * footprint * exposure_count
    dark_variance = dark_current * total_exposure_time * footprint

    total_variance = source_variance + sky_variance + readout_variance + dark_variance

    return np.sqrt(total_variance) * zp


def apply_noise(bandflux, bandflux_err, rng=None):
    """Apply Gaussian noise to a bandflux measurement.

    Parameters
    ----------
    bandflux : ndarray of float
        The bandflux measurement.
    bandflux_err : ndarray of float
        The bandflux measurement error.
    rng : np.random.Generator, optional
        The random number generator.

    Returns
    -------
    ndarray of float
        The noisy bandflux measurement.
    """
    if rng is None:
        rng = np.random.default_rng()

    return rng.normal(loc=bandflux, scale=bandflux_err)
